from collections import OrderedDict

import numpy as np
import pyqtgraph as pg
import pyqtgraph.parametertree.parameterTypes as pTypes
from PyQt5 import QtCore, QtGui, QtWidgets
from pyqtgraph import dockarea

from kite.covariance import CovarianceConfig
from kite.qt_utils import loadUi

from .base import KiteParameterGroup, KitePlot, KiteToolColormap, KiteView, get_resource

pen_covariance_model = pg.mkPen((204, 0, 0), width=2, style=QtCore.Qt.DotLine)

pen_covariance = pg.mkPen((255, 255, 255, 100), width=1.25)

pen_covariance_active = pg.mkPen((255, 255, 255), width=1.25)

pen_variance = pg.mkPen((78, 154, 6), width=2.5, style=QtCore.Qt.DashLine)
pen_variance_highlight = pg.mkPen((115, 210, 22), width=2.5, style=QtCore.Qt.DashLine)

pen_green_dash = pg.mkPen((115, 210, 22), width=2.5, style=QtCore.Qt.DashLine)

pen_roi = pg.mkPen((78, 154, 6), width=2)
pen_roi_highlight = pg.mkPen((115, 210, 22), width=2, style=QtCore.Qt.DashLine)


class KiteCovariance(KiteView):
    title = "Scene.covariance"

    def __init__(self, spool):
        model = spool.model
        self.model = model

        covariance_plot = KiteNoisePlot(model)
        self.main_widget = covariance_plot

        self.structure_function = KiteStructureFunction(covariance_plot)
        self.covariogram = KiteCovariogram(covariance_plot)

        self.tools = {
            # 'Covariance.powerspecNoise':
            #     KiteNoisePowerspec(covariance_plot),
            "Semi-Variogram: Covariance.structure_spatial": self.structure_function,
            "Covariogram: Covariance.covariance_spatial": self.covariogram,
        }

        self.param_covariance = KiteParamCovariance(model)
        self.parameters = [self.param_covariance]

        self.dialogInspectNoise = KiteToolNoise(model, spool)
        self.dialogInspectCovariance = KiteToolWeightMatrix(model, spool)
        self.dialogCovarianceInfo = CovarianceCalcResultDialog(model, spool)
        self.dialogCovarianceWarning = CovarianceNotPosDefWarning(model, spool)

        spool.actionInspect_Noise.triggered.connect(self.dialogInspectNoise.show)
        covariance_plot.roi.sigClicked.connect(self.dialogInspectNoise.show)
        spool.actionInspect_Weights.triggered.connect(self.dialogInspectCovariance.show)
        spool.actionCalculate_WeightMatrix.triggered.connect(
            lambda: QCalculateWeightMatrix(model, spool)
        )

        spool.actionInspect_Noise.setEnabled(True)
        spool.actionInspect_Weights.setEnabled(True)
        spool.actionCalculate_WeightMatrix.setEnabled(True)

        model.sigSceneModelChanged.connect(self.modelChanged)
        model.sigCalculateWeightMatrixFinished.connect(self.dialogCovarianceInfo.show)

        KiteView.__init__(self)

        for dock in self.tool_docks:
            dock.setStretch(10, 0.5)

    def modelChanged(self):
        self.dialogCovariance.close()
        self.main_widget.onConfigChanged()

        self.param_covariance.updateValues()
        for v in self.tools.values():
            v.update()

    @QtCore.pyqtSlot()
    def activateView(self):
        self.model.sigCovarianceChanged.connect(self.dialogCovarianceWarning.test)

        self.main_widget.update()
        self.structure_function.activatePlot()
        self.covariogram.activatePlot()

    @QtCore.pyqtSlot()
    def deactivateView(self):
        self.model.sigCovarianceChanged.disconnect(self.dialogCovarianceWarning.test)

        self.structure_function.deactivatePlot()
        self.covariogram.deactivatePlot()


class CovarianceNotPosDefWarning(QtWidgets.QMessageBox):
    def __init__(self, model, *args, **kwargs):
        QtWidgets.QMessageBox.__init__(self, *args, **kwargs)
        self.setIcon(QtWidgets.QMessageBox.Warning)

        self.setWindowTitle("Covariance Warning")
        self.setText(
            '<b><span style="font-family: monospace;">'
            "Covariance.covariance_matrix_focal</span>"
            " is not positive definite!</b>"
        )
        self.setInformativeText(
            'Change the <span style="font-family: monospace;">'
            "Covariance.model_function</span> to exponential<br>"
            "or move the noise patch to fix."
        )
        self.model = model

    @QtCore.pyqtSlot()
    def test(self):
        covariance = self.model.covariance
        if not covariance.isMatrixPosDefinite(full=False):
            self.show()


class KiteNoisePlot(KitePlot):
    class NoisePatchROI(pg.RectROI):
        def _makePen(self):
            # Generate the pen color for this ROI based on its current state.
            if self.mouseHovering:
                return pen_roi_highlight
            else:
                return self.pen

    def __init__(self, model):
        self.components_available = {
            "displacement": ["Displacement", lambda sp: sp.scene.displacement],
        }
        self._component = "displacement"

        KitePlot.__init__(self, model=model, los_arrow=True)

        llE, llN, sizeE, sizeN = self.model.covariance.noise_coord
        self.roi = self.NoisePatchROI(
            pos=(llE, llN), size=(sizeE, sizeN), sideScalers=True, pen=pen_roi
        )

        self.roi.setAcceptedMouseButtons(QtCore.Qt.RightButton)
        self.roi.sigRegionChangeFinished.connect(self.updateNoiseRegion)
        self.addItem(self.roi)

        self.model.sigCovarianceConfigChanged.connect(self.onConfigChanged)

    @QtCore.pyqtSlot()
    def onConfigChanged(self):
        llE, llN, sizeE, sizeN = self.model.covariance.noise_coord
        self.roi.setPos((llE, llN), update=False, finish=False)
        self.roi.setSize((sizeE, sizeN), finish=False)
        self.update()
        self.transFromFrame()

    @QtCore.pyqtSlot()
    def updateNoiseRegion(self):
        scene = self.model.getScene()

        llE, llN = self.roi.pos()
        sizeE, sizeN = self.roi.size()

        llEpx, llNpx = scene.frame.mapENMatrix(llE, llN)
        sEpx, sNpx = scene.frame.mapENMatrix(sizeE, sizeN)
        slice_E = slice(llEpx, llEpx + sEpx)
        slice_N = slice(llNpx, llNpx + sNpx)

        data = scene.displacement[slice_N, slice_E]
        data[data == 0.0] = np.nan
        if np.all(np.isnan(data)):
            return

        self.model.covariance.noise_coord = (llE, llN, sizeE, sizeN)
        self.model.covariance.noise_data = data


class KiteSubplot(QtWidgets.QWidget):
    def __init__(self, parent_plot):
        QtWidgets.QWidget.__init__(self)
        self.parent_plot = parent_plot
        self.model = parent_plot.model

        self.plot = pg.PlotWidget(background="default")
        self.plot.showGrid(True, True, alpha=0.5)
        self.plot.setMenuEnabled(True)
        self.plot.enableAutoRange()

        self.layout = QtWidgets.QVBoxLayout(self)
        self.layout.addWidget(self.plot)

    def addItem(self, *args, **kwargs):
        self.plot.addItem(*args, **kwargs)

    def activatePlot(self):
        pass

    def deactivatePlot(self):
        pass


class KiteNoisePowerspec(KiteSubplot):
    def __init__(self, parent_plot):
        KiteSubplot.__init__(self, parent_plot)

        self.power = pg.PlotDataItem(antialias=True)
        # self.power_lin = pg.PlotDataItem(antialias=True, pen=pen_green_dash)

        self.power.setZValue(10)
        self.plot.setLabels(
            bottom="Wavenumber (cycles/m)", left="Power (m<sup>2</sup>)"
        )

        self.plot.setLogMode(x=True, y=True)

        # self.legend = pg.LegendItem(offset=(0., .5))
        # self.legend.setParentItem(self.plot.graphicsItem())
        # self.legend.addItem(self.power_lin, 'Log-linear model')

        self.addItem(self.power)
        # self.addItem(self.power_lin)

    @QtCore.pyqtSlot()
    def update(self):
        covariance = self.model.covariance
        spec, k, _, _, _, _ = covariance.powerspecNoise1D()
        self.power.setData(k, spec)
        # self.power_lin.setData(
        #     k, covariance.powerspecModel(k))

    def activatePlot(self):
        self.model.sigCovarianceChanged.connect(self.update)
        self.update()

    def deactivatePlot(self):
        self.model.sigCovarianceChanged.disconnect(self.update)


class KiteCovariogram(KiteSubplot):

    legend_template = {
        "exponential": "Model: {0:.2g} e^(-d/{1:.1f}) | RMS: {rms:.4e}",
        "exponential_cosine": "Model: {0:.2g} e^(-d/{1:.1f}) - cos((d-({2:.1f}))/{3:.1f}) "
        "| RMS: {rms:.4e}",
    }

    class VarianceLine(pg.InfiniteLine):
        def __init__(self, *args, **kwargs):
            pg.InfiniteLine.__init__(self, *args, **kwargs)
            self.setCursor(QtCore.Qt.SizeVerCursor)

    def __init__(self, parent_plot):
        KiteSubplot.__init__(self, parent_plot)
        self.plot.setLabels(bottom=("Distance", "m"), left="Covariance (m<sup>2</sup>)")

        self.cov_spectral = pg.PlotDataItem(antialias=True)
        self.cov_spectral.setZValue(10)

        self.cov_spatial = pg.PlotDataItem(antialias=True)

        self.cov_model = pg.PlotDataItem(antialias=True, pen=pen_covariance_model)

        self.variance = self.VarianceLine(
            pen=pen_variance,
            angle=0,
            movable=True,
            hoverPen=pen_variance_highlight,
            label="Variance: {value:.5f}",
            labelOpts={
                "position": 0.975,
                "anchors": ((1.0, 0.0), (1.0, 1.0)),
                "color": pg.mkColor(255, 255, 255, 155),
            },
        )
        self.variance.setToolTip("Move to change variance")
        self.variance.sigPositionChangeFinished.connect(self.setVariance)

        self.addItem(self.cov_spectral)
        self.addItem(self.cov_spatial)
        self.addItem(self.cov_model)
        self.addItem(self.variance)
        # self.cov_lin_pow = pg.PlotDataItem(antialias=True,
        #                                    pen=pen_green_dash)
        # self.addItem(self.cov_lin_pow)

        self.legend = pg.LegendItem(offset=(0.0, 0.5))

        self.legend.setParentItem(self.plot.graphicsItem())
        self.legend.addItem(self.cov_model, "")

    def setVariance(self):
        self.model.covariance.variance = self.variance.value()

    @QtCore.pyqtSlot()
    def update(self):
        covariance = self.model.covariance

        cov_spectral, dist = covariance.covariance_spectral
        self.cov_spectral.setData(dist, cov_spectral)

        cov_spatial, dist = covariance.covariance_spatial
        self.cov_spatial.setData(dist, cov_spatial)

        if self.model.covariance.config.sampling_method == "spatial":
            self.cov_spatial.setPen(pen_covariance_active)
            self.cov_spectral.setPen(pen_covariance)

        else:
            self.cov_spatial.setPen(pen_covariance)
            self.cov_spectral.setPen(pen_covariance_active)

        model = self.model.covariance.getModelFunction()

        self.cov_model.setData(dist, model(dist, *covariance.covariance_model))

        tmpl = self.legend_template[covariance.config.model_function]

        self.legend.items[-1][1].setText(
            tmpl.format(
                *covariance.covariance_model, rms=covariance.covariance_model_rms
            )
        )
        self.variance.setValue(covariance.variance)

    def activatePlot(self):
        self.model.sigCovarianceChanged.connect(self.update)
        self.update()

    def deactivatePlot(self):
        self.model.sigCovarianceChanged.disconnect(self.update)


class KiteStructureFunction(KiteSubplot):
    class VarianceLine(pg.InfiniteLine):
        def __init__(self, *args, **kwargs):
            pg.InfiniteLine.__init__(self, *args, **kwargs)
            self.setCursor(QtCore.Qt.SizeVerCursor)

    def __init__(self, parent_plot):
        KiteSubplot.__init__(self, parent_plot)

        self.structure = pg.PlotDataItem(antialias=True, pen=pen_covariance_active)
        self.variance = self.VarianceLine(
            pen=pen_variance,
            angle=0,
            movable=True,
            hoverPen=pen_variance_highlight,
            label="Variance: {value:.5f}",
            labelOpts={
                "position": 0.975,
                "anchors": ((1.0, 0.0), (1.0, 1.0)),
                "color": pg.mkColor(255, 255, 255, 155),
            },
        )
        self.plot.setLabels(bottom=("Distance", "m"), left="Variance (m<sup>2</sup>)")

        self.addItem(self.structure)
        self.addItem(self.variance)
        self.variance.sigPositionChangeFinished.connect(self.changeVariance)

    @QtCore.pyqtSlot()
    def update(self):
        covariance = self.model.covariance
        struct, dist = covariance.getStructure()
        self.structure.setData(dist[np.isfinite(struct)], struct[np.isfinite(struct)])
        self.variance.setValue(covariance.variance)

    def changeVariance(self, inf_line):
        covariance = self.model.covariance
        covariance.variance = inf_line.getYPos()

    def activatePlot(self):
        self.model.sigCovarianceChanged.connect(self.update)
        self.update()

    def deactivatePlot(self):
        self.model.sigCovarianceChanged.disconnect(self.update)


class KiteToolNoise(QtWidgets.QDialog):
    class NoisePlot(KitePlot):
        def __init__(self, model):
            self.components_available = {
                "noise_data": [
                    "Displacement",
                    lambda sp: self.noise_data_masked(sp.covariance),
                ]
            }

            self._component = "noise_data"
            KitePlot.__init__(self, model=model)

        @staticmethod
        def noise_data_masked(covariance):
            data = covariance.noise_data.copy()
            data[data == 0.0] = np.nan
            return data

        def proxy_connect(self):
            self.model.sigCovarianceChanged.connect(self.update)

        @QtCore.pyqtSlot()
        def update(self):
            KitePlot.update(self)

        def proxy_disconnect(self):
            self.model.sigCovarianceChanged.disconnect(self.update)

    class NoiseSyntheticPlot(KitePlot):
        def __init__(self, model):
            sp = model
            _, _, sizeE, sizeN = sp.covariance.noise_coord

            self.patch_size_roi = pg.RectROI(
                pos=(0.0, 0.0),
                size=(sizeE, sizeN),
                sideScalers=True,
                movable=False,
                pen=pen_roi,
            )

            self._anisotropic = False
            self.components_available = {
                "synthetic_noise": [
                    "Noise",
                    lambda sp: sp.covariance.syntheticNoise(
                        self.sizePatchPx(), anisotropic=self.anisotropic
                    ),
                ]
            }

            self._component = "synthetic_noise"

            KitePlot.__init__(self, model=model)
            self.patch_size_roi.sigRegionChangeFinished.connect(self.update)
            self.addItem(self.patch_size_roi)

        @property
        def anisotropic(self):
            return self._anisotropic

        def enableAnisotropic(self, value):
            self._anisotropic = value
            self.update()

        def resetSize(self):
            _, _, sizeE, sizeN = self.model.covariance.noise_coord
            self.patch_size_roi.setSize((sizeE, sizeN))

        def sizePatchPx(self):
            sp = self.model
            sizeE, sizeN = self.patch_size_roi.size()
            return int(sizeN / sp.frame.dN), int(sizeE / sp.frame.dE)

        def setGradientEditor(self, gradient_editor):
            ge = gradient_editor
            image = self.image

            hist_pen = pg.mkPen((170, 57, 57, 255), width=1.0)
            image.setLookupTable(ge.getLookupTable)

            def updateLevels():
                image.setLevels(ge.region.getRegion())

            ge.sigLevelChangeFinished.connect(updateLevels)
            ge.sigLevelsChanged.connect(updateLevels)
            updateLevels()

            def updateHistogram():
                h = image.getHistogram()
                if h[0] is None:
                    return
                ge.hist_syn.setData(*h)

            ge.hist_syn = pg.PlotDataItem(pen=hist_pen)
            ge.hist_syn.rotate(90.0)
            ge.vb.addItem(ge.hist_syn)
            updateHistogram()

            image.sigImageChanged.connect(updateHistogram)

        def proxy_connect(self):
            self.model.sigCovarianceChanged.connect(self.update)

        @QtCore.pyqtSlot()
        def update(self):
            KitePlot.update(self)

        def proxy_disconnect(self):
            self.model.sigCovarianceChanged.disconnect(self.update)

    def __init__(self, model, parent=None):
        QtWidgets.QDialog.__init__(self, parent)

        loadUi(get_resource("noise_dialog.ui"), baseinstance=self)
        self.closeButton.setIcon(
            self.style().standardIcon(QtWidgets.QStyle.SP_DialogCloseButton)
        )
        self.setWindowFlags(QtCore.Qt.Window)

        self.noise_patch = self.NoisePlot(model)
        self.noise_synthetic = self.NoiseSyntheticPlot(model)

        colormap = KiteToolColormap(self.noise_patch)
        self.noise_synthetic.setGradientEditor(colormap)

        self.dockarea = dockarea.DockArea(self)

        self.dockarea.addDock(
            dockarea.Dock(
                "Covariance.noise_data",
                widget=self.noise_patch,
                size=(6, 6),
                autoOrientation=False,
            ),
            position="top",
        )

        self.dockarea.addDock(
            dockarea.Dock(
                "Covariance.syntheticNoise",
                widget=self.noise_synthetic,
                size=(6, 6),
                autoOrientation=False,
            ),
            position="bottom",
        )

        self.dockarea.addDock(
            dockarea.Dock(
                "Colormap",
                widget=colormap,
                size=(1, 1),
                autoOrientation=False,
            ),
            position="right",
        )
        self.horizontalLayoutPlot.addWidget(self.dockarea)

        self.resetSizeButton.released.connect(self.noise_synthetic.resetSize)
        self.anisotropicCB.toggled.connect(
            lambda b: self.noise_synthetic.enableAnisotropic(b)
        )

    def closeEvent(self, ev):
        self.noise_patch.proxy_disconnect()
        self.noise_synthetic.proxy_disconnect()
        ev.accept()

    def showEvent(self, ev):
        self.noise_patch.update()
        self.noise_patch.proxy_connect()
        self.noise_synthetic.update()
        self.noise_synthetic.proxy_connect()
        ev.accept()


class KiteToolWeightMatrix(QtWidgets.QDialog):
    class MatrixPlot(KitePlot):
        def __init__(self, model, parent):
            from pyqtgraph.graphicsItems.GradientEditorItem import Gradients

            self._component = "weight"
            self.parent = parent

            KitePlot.__init__(self, model)
            self.model = model

            gradient = Gradients["thermal"]

            self.cmap = pg.ColorMap(
                pos=[c[0] for c in gradient["ticks"]],
                color=[c[1] for c in gradient["ticks"]],
                mode=gradient["mode"],
            )
            self.image.setLookupTable(self.cmap.getLookupTable())

            self.setLabels(bottom={"Leaf #", ""}, left={"Leaf #", ""})

            self.setAspectLocked(True)
            self.setMouseEnabled(x=False, y=False)

            self.hint = {"leaf1": 0, "leaf2": 0, "weight": np.nan}

            self.hint_text.template = (
                '<span style="font-family: monospace; color: #fff;'
                'background-color: #000;">'
                "Leaf #1: {leaf1:d} | Leaf #2: {leaf2:d} | "
                "{weight:e}</span>"
            )

            self.update()

        @QtCore.pyqtSlot()
        def update(self):
            button = self.parent.matrixButtonGroup.checkedButton()
            checked_matrix = button.text()
            if checked_matrix == "covariance_matrix":
                matrix = self.model.covariance.covariance_matrix
            elif checked_matrix == "covariance_matrix_focal":
                matrix = self.model.covariance.covariance_matrix_focal
            elif checked_matrix == "weight_matrix_focal":
                matrix = self.model.covariance.weight_matrix_focal
            elif checked_matrix == "weight_matrix":
                matrix = self.model.covariance.weight_matrix

            self.image.updateImage(matrix.T, autoLevels=True)
            self.transFromFrame()

        def transFromFrame(self):
            # self.resetTransform()
            self.setRange(
                xRange=(0, self.model.quadtree.nleaves),
                yRange=(0, self.model.quadtree.nleaves),
            )

        @QtCore.pyqtSlot(object)
        def mouseMoved(self, event=None):
            if event is None:
                pass
            elif self.image.sceneBoundingRect().contains(event[0]):
                map_pos = self.plotItem.vb.mapSceneToView(event[0])
                if not map_pos.isNull():
                    img_pos = self.image.mapFromScene(*event)
                    value = self.image.image[int(img_pos.x()), int(img_pos.y())]

                    self.hint["leaf1"] = int(map_pos.x())
                    self.hint["leaf2"] = int(map_pos.y())
                    self.hint["weight"] = value
            self.hint_text.setText(self.hint_text.template.format(**self.hint))

        def proxy_connect(self):
            self.model.sigCovarianceChanged.connect(self.update)
            self.parent.matrixButtonGroup.buttonClicked.connect(self.update)

        def proxy_disconnect(self):
            self.model.sigCovarianceChanged.disconnect(self.update)
            self.parent.matrixButtonGroup.buttonClicked.disconnect(self.update)

    def __init__(self, model, parent=None):
        QtWidgets.QDialog.__init__(self, parent)
        self.model = model

        loadUi(get_resource("weight_matrix.ui"), baseinstance=self)
        self.closeButton.setIcon(
            self.style().standardIcon(QtWidgets.QStyle.SP_DialogCloseButton)
        )

        self.weight_matrix = self.MatrixPlot(model, self)
        self.dockarea = dockarea.DockArea(self)

        self.dockarea.addDock(
            dockarea.Dock(
                "Covariance.weight_matrix_focal",
                widget=self.weight_matrix,
                size=(4, 4),
                autoOrientation=False,
            ),
            position="left",
        )

        self.horizontalLayoutPlot.addWidget(self.dockarea)
        self.model.sigCovarianceChanged.connect(self.updateMatrixButtons)
        self.model.sigCovarianceConfigChanged.connect(self.updateMatrixButtons)

    @QtCore.pyqtSlot()
    def updateMatrixButtons(self):
        if self.model.covariance.isFullCovarianceCalculated():
            self.showCovarianceFull.setEnabled(True)
            self.showWeightFull.setEnabled(True)
        else:
            self.showCovarianceFull.setEnabled(False)
            self.showWeightFull.setEnabled(False)
            if self.showWeightFull.isChecked() or self.showCovarianceFull.isChecked():
                self.showWeightFocal.setChecked(True)

    def closeEvent(self, ev):
        self.weight_matrix.proxy_disconnect()
        ev.accept()

    def showEvent(self, ev):
        self.weight_matrix.proxy_connect()
        self.weight_matrix.update()
        self.updateMatrixButtons()
        ev.accept()


class QCalculateWeightMatrix(QtCore.QObject):
    sigCalculateWeightMatrix = QtCore.pyqtSignal()

    def __init__(self, model, parent):
        QtCore.QObject.__init__(self)
        self.sigCalculateWeightMatrix.connect(model.calculateWeightMatrix)

        diag = QtWidgets.QMessageBox.information(
            parent,
            "Calculate Full Weight Matrix",
            """<html><head/><body><p>
This will calculate the quadtree's full weight matrix
(<span style='font-family: monospace'>Covariance.weight_matrix</span>)
for this noise/covariance configuration.</p><p>
The calculation is expensive and may take several minutes.
</p></body></html>
""",
            buttons=(QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel),
        )

        if diag == QtWidgets.QMessageBox.Ok:
            meta = model.metaObject()
            meta.invokeMethod(
                model, "calculateWeightMatrix", QtCore.Qt.QueuedConnection
            )


class CovarianceCalcResultDialog(QtWidgets.QMessageBox):

    text_tmpl = (
        '<span style="font-family: monospace;">' "Covariance.covariance_matrix</span>"
    )

    def __init__(self, model, *args, **kwargs):
        QtWidgets.QMessageBox.__init__(self, *args, **kwargs)

        self.setIcon(QtWidgets.QMessageBox.Information)
        self.model = model
        self.setWindowTitle("Covariance Calculation")
        self.setTextFormat(QtCore.Qt.RichText)

    @QtCore.pyqtSlot(object)
    def show(self, elapsed_time, *args, **kwargs):

        if self.model.covariance.isMatrixPosDefinite(full=True):
            self.setIcon(QtWidgets.QMessageBox.Information)
            self.setText("Finished, %s is positive definite!" % self.text_tmpl)
            self.setInformativeText("")
        else:
            self.setIcon(QtWidgets.QMessageBox.Warning)
            self.setText("<b>%s is not positive definite!</b>" % self.text_tmpl)
            self.setInformativeText(
                'Change the <span style="font-family: monospace;">'
                "Covariance.model_function</span> to exponential"
                " or moving the noise patch."
            )

        self.setDetailedText("Elapsed time: %s" % elapsed_time)
        QtWidgets.QMessageBox.show(self, *args, **kwargs)


class KiteParamCovariance(KiteParameterGroup):
    sigSamplingMethod = QtCore.pyqtSignal(str)
    sigSpatialBins = QtCore.pyqtSignal(int)
    sigSpatialPairs = QtCore.pyqtSignal(int)

    def __init__(self, model, **kwargs):
        kwargs["type"] = "group"
        kwargs["name"] = "Scene.covariance"

        self.sp = model

        self.parameters = OrderedDict(
            [
                ("variance", None),
                (
                    "covariance_model",
                    lambda c: ", ".join("%g" % p for p in c.covariance_model),
                ),
                ("covariance_model_rms", None),
                ("noise_patch_size_km2", None),
                (
                    "noise_patch_coord",
                    lambda c: ", ".join(["%.2f" % f for f in c.noise_coord.tolist()]),
                ),
            ]
        )

        model.sigCovarianceChanged.connect(self.updateValues)
        KiteParameterGroup.__init__(
            self, model=model, model_attr="covariance", **kwargs
        )

        def changeSamplingMethod():
            model.covariance.setSamplingMethod(sampling_method.value())

        p = {
            "name": "sampling_method",
            "values": {
                "spatial random": "spatial",
                "spectral": "spectral",
            },
            "value": model.covariance.config.sampling_method,
            "tip": CovarianceConfig.sampling_method.help,
        }
        sampling_method = pTypes.ListParameter(**p)
        sampling_method.sigValueChanged.connect(changeSamplingMethod)

        def changeSpatialBins():
            model.covariance.setSpatialBins(spatial_bins.value())

        p = {
            "name": "spatial_bins",
            "value": model.covariance.config.spatial_bins,
            "type": "int",
            "limits": (1, 500),
            "step": 5,
            "editable": True,
            "tip": CovarianceConfig.spatial_bins.help,
        }

        spatial_bins = pTypes.SimpleParameter(**p)
        spatial_bins.sigValueChanged.connect(changeSpatialBins)

        def changeSpatialPairs():
            model.covariance.setSpatialPairs(spatial_pairs.value())

        p = {
            "name": "spatial_pairs",
            "value": model.covariance.config.spatial_pairs,
            "type": "int",
            "limits": (1, 1000000),
            "step": 50000,
            "editable": True,
            "tip": CovarianceConfig.spatial_pairs.help,
        }

        spatial_pairs = pTypes.SimpleParameter(**p)
        spatial_pairs.sigValueChanged.connect(changeSpatialPairs)

        def changeModelFunction():
            model.covariance.setModelFunction(model_function.value())

        p = {
            "name": "model_function",
            "values": {
                "exponential": "exponential",
                "exp + cosine": "exponential_cosine",
            },
            "value": model.covariance.config.model_function,
            "tip": CovarianceConfig.model_function.help,
        }
        model_function = pTypes.ListParameter(**p)
        model_function.sigValueChanged.connect(changeModelFunction)

        def toggle_adaptive_subsampling(param, checked):
            model.covariance.config.adaptive_subsampling = checked

        p = {
            "name": "adaptive_subsampling",
            "type": "bool",
            "value": model.covariance.config.adaptive_subsampling,
            "tip": "detrend the scene",
        }
        adaptive_subsampling = pTypes.SimpleParameter(**p)
        adaptive_subsampling.sigValueChanged.connect(toggle_adaptive_subsampling)

        self.pushChild(adaptive_subsampling)
        self.pushChild(model_function)
        self.pushChild(spatial_bins)
        self.pushChild(spatial_pairs)
        self.pushChild(sampling_method)



import numpy as np
from pyrocko import orthodrome as od
from pyrocko.guts import Float, Object

from kite.util import Subject

d2r = np.pi / 180.0
r2d = 180.0 / np.pi
km = 1e3


class SandboxSource(Object):

    lat = Float.T(default=0.0, help="Latitude in [deg]")
    lon = Float.T(default=0.0, help="Longitude in [deg]")
    easting = Float.T(default=0.0, help="Easting in [m]")
    northing = Float.T(default=0.0, help="Northing in [m]")
    depth = Float.T(default=1.0 * km, help="Depth in [m]")

    def __init__(self, *args, **kwargs):
        Object.__init__(self, *args, **kwargs)
        self._cached_result = None
        self.evParametersChanged = Subject()
        self._sandbox = None

    def parametersUpdated(self):
        self._cached_result = None
        self.evParametersChanged.notify()

    def getSandboxOffset(self):
        if not self._sandbox or (self.lat == 0.0 and self.lon == 0.0):
            return 0.0, 0.0
        return od.latlon_to_ne_numpy(
            self._sandbox.frame.llLat, self._sandbox.frame.llLon, self.lat, self.lon
        )

    def getParametersArray(self):
        raise NotImplementedError


class SandboxSourceRectangular(SandboxSource):

    width = Float.T(
        help="Width, downdip in [m]",
        default=10000.0,
    )
    length = Float.T(
        help="Length in [m]",
        default=10000.0,
    )
    strike = Float.T(
        default=45.0, help="Strike, clockwise from North in [deg]; -180-180"
    )
    dip = Float.T(default=45.0, help="Dip, down from horizontal in [deg]; 0-90")
    rake = Float.T(
        default=90.0, help="Rake, clockwise in [deg]; 0 is right-lateral strike slip"
    )
    slip = Float.T(default=1.5, help="Slip in [m]", optional=True)

    def outline(self):
        coords = np.empty((4, 2))

        c_strike = np.cos(self.strike * d2r)
        s_strike = np.sin(self.strike * d2r)
        c_dip = np.cos(self.dip * d2r)

        coords[0, 0] = s_strike * self.length / 2
        coords[0, 1] = c_strike * self.length / 2
        coords[1, 0] = -coords[0, 0]
        coords[1, 1] = -coords[0, 1]

        coords[2, 0] = coords[1, 0] - c_strike * c_dip * self.width
        coords[2, 1] = coords[1, 1] + s_strike * c_dip * self.width
        coords[3, 0] = coords[0, 0] - c_strike * c_dip * self.width
        coords[3, 1] = coords[0, 1] + s_strike * c_dip * self.width

        north_shift, east_shift = self.getSandboxOffset()

        coords[:, 0] += self.easting + east_shift
        coords[:, 1] += self.northing + north_shift

        return coords

    @property
    def segments(self):
        yield self

    @classmethod
    def fromPyrockoSource(cls, source, store_dir=None, **kwargs):
        d = dict(
            lat=source.lat,
            lon=source.lon,
            northing=source.north_shift,
            easting=source.east_shift,
            depth=source.depth,
            width=source.width,
            length=source.length,
            strike=source.strike,
            dip=source.dip,
            rake=source.rake,
            slip=source.slip,
        )

        if hasattr(cls, "decimation_factor"):
            d["decimation_factor"] = source.decimation_factor

        if hasattr(cls, "store_dir"):
            d["store_dir"] = store_dir

        return cls(**d)


class ProcessorProfile(dict):
    pass


class SourceProcessor(object):
    """Interface definition of the processor"""

    __implements__ = "ProcessorName"  # Defines which backend is implemented

    def __init__(self, sandbox_scene):
        self._log = sandbox_scene._log.getChild(self.__class__.__name__)

    def process(sources, coords, nthreads=0):
        raise NotImplementedError()

        result = {
            "displacement.n": np.array(),
            "displacement.e": np.array(),
            "displacement.d": np.array(),
            "processor_profile": ProcessorProfile(),
        }
        return result



import logging

from PyQt5 import QtCore

# Importing available models
from kite.qt_utils import SceneLogModel

from . import sources
from .util import SourceEditorDialog, SourceROI

available_delegates = {}
for module in sources.__sources__:
    available_delegates[module.__represents__] = module


class CursorTracker(QtCore.QObject):
    sigCursorMoved = QtCore.pyqtSignal(object)
    sigMouseMoved = QtCore.pyqtSignal(object)


class SandboxModel(QtCore.QObject):

    sigModelUpdated = QtCore.pyqtSignal()
    sigModelChanged = QtCore.pyqtSignal()
    sigLogRecord = QtCore.pyqtSignal(object)

    sigProcessingFinished = QtCore.pyqtSignal()
    sigProcessingStarted = QtCore.pyqtSignal(str)

    def __init__(self, sandbox_model=None, *args, **kwargs):
        QtCore.QObject.__init__(self)

        self.model = sandbox_model
        self.log = SceneLogModel(self)
        self.sources = SourceModel(self)
        self.cursor_tracker = CursorTracker(self)

        self._log_handler = logging.Handler()
        self._log_handler.setLevel(logging.DEBUG)
        self._log_handler.emit = self.sigLogRecord.emit

        logging.root.setLevel(logging.DEBUG)
        logging.root.addHandler(self._log_handler)

        self.worker_thread = QtCore.QThread(self)
        self.moveToThread(self.worker_thread)
        self.worker_thread.start()

        if self.model:
            self.setModel(self.model)

    def setModel(self, model):
        self.disconnectSlots()

        self.model = model
        self.frame = model.frame

        self.connectSlots()
        self.sigModelChanged.emit()
        self.sigModelUpdated.emit()

    def connectSlots(self):
        self.model._log.addHandler(self._log_handler)
        self.model.evModelUpdated.subscribe(self.sigModelUpdated.emit)

    def disconnectSlots(self):
        if self.model is None:
            return
        try:
            self.model._log.removeHandler(self._log_handler)
            self.model.evModelUpdated.unsubscribe(self.sigModelUpdated.emit)
        except AttributeError:
            pass
        finally:
            self.model = None

    def addSource(self, source):
        self.model.addSource(source)

    def removeSource(self, source):
        self.model.removeSource(source)

    @QtCore.pyqtSlot()
    def optimizeSource(self):
        self.sigProcessingStarted.emit("Optimizing source, stay tuned!")
        self.model.reference.optimizeSource()
        self.sigProcessingFinished.emit()

    @classmethod
    def randomOkada(cls, nsources=1):
        from ..sandbox_scene import TestSandboxScene

        model = TestSandboxScene.randomOkada(nsources)
        sandbox = cls(model)
        return sandbox

    @classmethod
    def simpleOkada(cls, **kwargs):
        from ..sandbox_scene import TestSandboxScene

        model = TestSandboxScene.simpleOkada(**kwargs)
        sandbox = cls(model)
        return sandbox

    @classmethod
    def empty(cls, parent=None):
        from ..sandbox_scene import SandboxScene

        sandbox = cls(parent=parent)
        sandbox.setModel(SandboxScene())
        return sandbox


class SourceModel(QtCore.QAbstractTableModel):

    selectionModelChanged = QtCore.pyqtSignal()

    def __init__(self, sandbox, *args, **kwargs):
        QtCore.QAbstractTableModel.__init__(self, *args, **kwargs)

        self.sandbox = sandbox
        self.selection_model = None
        self._createSources()

        self.sandbox.sigModelUpdated.connect(self.modelUpdated)
        self.sandbox.sigModelChanged.connect(self.modelChanged)

    def _createSources(self):
        self._sources = []
        for isrc, src in enumerate(self.model_sources):
            source_model = available_delegates[src.__class__.__name__]
            idx = self.createIndex(isrc, 0)
            src = source_model(self, src, idx, parent=self)

            self._sources.append(src)

    @property
    def model_sources(self):
        if self.sandbox.model is None:
            return []
        else:
            return self.sandbox.model.sources

    def rowCount(self, idx):
        return len(self.model_sources)

    def columnCount(self, idx):
        return 1

    def flags(self, idx):
        return (
            QtCore.Qt.ItemIsSelectable
            | QtCore.Qt.ItemIsEditable
            | QtCore.Qt.ItemIsEnabled
        )

    def setSelectionModel(self, selection_model):
        self.selection_model = selection_model
        self.selectionModelChanged.emit()

    def data(self, idx, role):
        src = self._sources[idx.row()]
        if role == QtCore.Qt.DisplayRole:
            return src.formatListItem()
        elif role == SourceROI:
            return src.getROIItem()
        elif role == SourceEditorDialog:
            return src.getEditingDialog()

    def itemData(self, idx):
        src = self._sources[idx.row()]
        return src.getSourceParameters()

    def setItemData(self, idx, parameters):
        src = self._sources[idx.row()]
        src.setSourceParameters(parameters)
        self.dataChanged.emit(idx, idx)
        return True

    def setData(self, idx, value, role):
        print("Set %s with role %s to value %s" % (idx, value, role))

    def removeSource(self, idx):
        src = self._sources[idx.row()]
        self.sandbox.removeSource(src.source)

    @QtCore.pyqtSlot()
    def modelUpdated(self, force=False):
        if len(self._sources) != len(self.model_sources) or force:
            self.beginResetModel()
            self._createSources()
            self.endResetModel()

    @QtCore.pyqtSlot()
    def modelChanged(self):
        self.modelUpdated(force=True)



#!/bin/python
import logging
import unittest

import numpy as np

from kite import Scene, TestScene

# from . import common

# common.setLogLevel('DEBUG')


def get_scene():
    sc = Scene()
    sc.frame.llLat = 52.395833
    sc.frame.llLon = 13.061389
    sc.frame.dE = 0.001
    sc.frame.dN = 0.001
    sc.frame.spacing = "degree"
    sc.displacement = np.zeros((500, 500))

    return sc


class TestSyntheticsGenerators(unittest.TestCase):
    def testGauss(self):
        TestScene.createGauss()

    def testFractal(self):
        TestScene.createFractal()

    def testSine(self):
        TestScene.createSine()


class TestSyntheticScene(unittest.TestCase):
    def setUp(self):
        self.sc = TestScene.createGauss()
        self.sc.quadtree.epsilon = 0.02
        self.sc.covariance.subsampling = 24

    def testQuadtree(self):
        qt = self.sc.quadtree
        for e in np.linspace(0.118, 0.2, num=30):
            qt.epsilon = e

        for nan in np.linspace(0.1, 1.0, num=30):
            qt.nan_allowed = nan

        for s in np.linspace(100, 4000, num=30):
            qt.tile_size_min = s
            qt.tile_size_max = 5000

        for s in np.linspace(200, 4000, num=30):
            qt.tile_size_min = 20
            qt.tile_size_max = s


def test_topo():
    logging.basicConfig(level=logging.DEBUG)
    sc = get_scene()
    return sc.get_elevation()


if __name__ == "__main__":
    test_topo()



import numpy as np
from PyQt5 import QtCore, QtWidgets

from kite.qt_utils import loadUi

from .util import get_resource

km = 1e3


class ExtentDialog(QtWidgets.QDialog):
    def __init__(self, sandbox, *args, **kwargs):
        super().__init__(*args, **kwargs)
        loadUi(get_resource("dialog_extent.ui"), self)
        self.setSizeGripEnabled(False)

        self.move(
            self.parent().window().mapToGlobal(self.parent().window().rect().center())
            - self.mapToGlobal(self.rect().center())
        )

        self.sandbox = sandbox
        model = self.sandbox.model
        dE, dN = model.frame.dEmeter, model.frame.dNmeter

        def getKm(px, dp):
            return "%.2f km " % (dp * px / km)

        self.spinEastPx.valueChanged.connect(
            lambda px: self.eastKm.setText(getKm(px, dE))
        )
        self.spinNorthPx.valueChanged.connect(
            lambda px: self.northKm.setText(getKm(px, dN))
        )

        self.applyButton.released.connect(self.updateValues)
        self.okButton.released.connect(self.updateValues)
        self.okButton.released.connect(self.close)

        self.setValues()

    def setValues(self, east=None, north=None):
        model = self.sandbox.model

        east = model.config.extent_east if east is None else east
        north = model.config.extent_north if north is None else north

        self.spinEastPx.setValue(east)
        self.spinNorthPx.setValue(north)

    @QtCore.pyqtSlot()
    def updateValues(self):
        self.sandbox.model.setExtent(self.spinEastPx.value(), self.spinNorthPx.value())
        self.setValues(self.spinEastPx.value(), self.spinEastPx.value())
        self.setValues()


class LosDialog(QtWidgets.QDialog):
    def __init__(self, sandbox, *args, **kwargs):
        QtWidgets.QDialog.__init__(self, *args, **kwargs)
        loadUi(get_resource("dialog_los.ui"), self)
        self.setSizeGripEnabled(False)

        self.move(
            self.parent().window().mapToGlobal(self.parent().window().rect().center())
            - self.mapToGlobal(self.rect().center())
        )

        self.sandbox = sandbox
        model = self.sandbox.model
        self.applyButton.released.connect(self.updateValues)
        self.okButton.released.connect(self.updateValues)
        self.okButton.released.connect(self.close)

        self.setValues()

    def setValues(self):
        model = self.sandbox.model
        phi = np.deg2rad(self.spinlos_phi.value())
        theta = np.deg2rad(self.spinlos_theta.value())

    @QtCore.pyqtSlot()
    def updateValues(self):
        print("updated los!")
        self.sandbox.model.setLOS(self.spinlos_phi.value(), self.spinlos_theta.value())
        self.setValues()



import numpy as np
from pyrocko import orthodrome as od
from pyrocko.guts import Bool, Float, List

from kite.sources import disloc_ext

from .base import SandboxSource, SandboxSourceRectangular, SourceProcessor

d2r = np.pi / 180.0
r2d = 180.0 / np.pi
km = 1e3


__all__ = ["OkadaSource", "OkadaPath", "DislocProcessor"]


class OkadaSource(SandboxSourceRectangular):
    """Rectangular Okada source model."""

    __implements__ = "disloc"

    opening = Float.T(help="Opening of the plane in [m]", optional=True, default=0.0)
    nu = Float.T(default=0.25, help="Poisson's ratio, typically 0.25")

    @property
    def seismic_moment(self):
        """Scalar Seismic moment

        Disregarding the opening (as for now)
        We assume a shear modulus of :math:`\mu = 36 \mathrm{GPa}`
        and :math:`M_0 = \mu A D`

        .. important ::

            We assume a perfect elastic solid with :math:`K=\\frac{5}{3}\\mu`

            Through :math:`\\mu = \\frac{3K(1-2\\nu)}{2(1+\\nu)}` this leads to
            :math:`\\mu = \\frac{8(1+\\nu)}{1-2\\nu}`

        :returns: Seismic moment release
        :rtype: float
        """
        mu = (8.0 * (1 + self.nu)) / (1 - 2.0 * self.nu)
        mu = 32e9  # GPa
        A = self.length * self.width
        return mu * A * self.slip

    @property
    def moment_magnitude(self):
        """Moment magnitude from Seismic moment

        We assume :math:`M_\\mathrm{w} = {\\frac{2}{3}}\\log_{10}(M_0) - 10.7`

        :returns: Moment magnitude
        :rtype: float
        """
        return 2.0 / 3 * np.log10(self.seismic_moment * 1e7) - 10.7

    def dislocSource(self, dsrc=None):
        if dsrc is None:
            dsrc = np.empty(10)

        dip = self.dip
        if self.dip == 90.0:
            dip -= 1e-2

        dsrc[0] = self.length
        dsrc[1] = self.width
        dsrc[2] = self.depth
        dsrc[3] = -dip
        dsrc[4] = self.strike - 180.0
        dsrc[5] = self.easting
        dsrc[6] = self.northing

        ss_slip = np.cos(self.rake * d2r) * self.slip
        ds_slip = np.sin(self.rake * d2r) * self.slip
        # print '{:<13}{}\n{:<13}{}'.format(
        #     'strike_slip', ss_slip, 'dip_slip', ds_slip)
        dsrc[7] = -ss_slip  # SS Strike-Slip
        dsrc[8] = -ds_slip  # DS Dip-Slip
        dsrc[9] = self.opening  # TS Tensional-Slip

        return dsrc

    # @property
    # def parameters(self):
    # return self.T.propnames

    def getParametersArray(self):
        return np.array([self.__getattribute__(p) for p in self.parameters])

    def setParametersArray(self, parameter_arr):
        if parameter_arr.size != len(self.parameters):
            raise AttributeError(
                "Invalid number of parameters, %s has %d" " parameters" % self.__name__,
                len(self.parameters),
            )
        for ip, param in enumerate(self.parameters):
            self.__setattr__(param, parameter_arr[ip])
        self.parametersUpdated()


class OkadaSegment(OkadaSource):
    enabled = Bool.T(default=True, optional=True)


class OkadaPath(SandboxSource):

    __implements__ = "disloc"

    depth = None
    nu = Float.T(default=0.25, help="Poisson's ratio, typically 0.25")
    nodes = List.T(
        default=[],
        optional=True,
        help="Nodes of the segments as (easting, northing) tuple of [m]",
    )
    segments__ = List.T(default=[], optional=True, help="List of all segments.")

    def __init__(self, *args, **kwargs):
        SandboxSource.__init__(self, *args, **kwargs)

        self._segments = []

        if not self.nodes:
            self.nodes.append([self.easting, self.northing])

    @property
    def segments(self):
        return self._segments

    @segments.setter
    def segments(self, segments):
        self._segments = segments

    @staticmethod
    def _newSegment(e1, n1, e2, n2, **kwargs):
        dE = e2 - e1
        dN = n2 - n1
        length = (dN**2 + dE**2) ** 0.5
        """Width Scaling relation after

        Leonard, M. (2010). Earthquake fault scaling: Relating rupture length,
            width, average displacement, and moment release, Bull. Seismol.
            Soc. Am. 100, no. 5, 1971-1988.
        """
        segment = {
            "northing": n1 + dN / 2,
            "easting": e1 + dE / 2,
            "depth": 0.0,
            "length": length,
            "width": 15.0 * length**0.66,
            "strike": np.arccos(dN / length) * r2d,
            "slip": 45.0,
            "rake": 90.0,
        }
        segment.update(kwargs)
        return OkadaSegment(**segment)

    def _moveSegment(self, pos, e1, n1, e2, n2):
        dE = e2 - e1
        dN = n2 - n1
        length = (dN**2 + dE**2) ** 0.5

        segment_update = {
            "northing": n1 + dN / 2,
            "easting": e1 + dE / 2,
            "length": length,
            "width": 15.0 * length**0.66,
            "strike": np.arccos(dN / length) * r2d,
        }

        segment = self.segments[pos]
        for attr, val in segment_update.items():
            segment.__setattr__(attr, val)

    def addNode(self, easting, northing):
        self.nodes.append([easting, northing])
        self.segments.append(
            self._newSegment(
                e1=self.nodes[-2][0],
                n1=self.nodes[-2][1],
                e2=self.nodes[-1][0],
                n2=self.nodes[-1][1],
            )
        )

    def insertNode(self, pos, easting, northing):
        self.nodes.insert(pos, [easting, northing])
        self.segments.append(
            self._newSegment(
                e1=self.nodes[pos][0],
                n1=self.nodes[pos][1],
                e2=self.nodes[pos + 1][0],
                n2=self.nodes[pos + 1][1],
            )
        )
        self._moveSegment(
            pos - 1,
            e1=self.nodes[pos - 1][0],
            n1=self.nodes[pos - 1][1],
            e2=self.nodes[pos][0],
            n2=self.nodes[pos][1],
        )

    def moveNode(self, pos, easting, northing):
        self.nodes[pos] = [easting, northing]
        if pos < len(self):
            self._moveSegment(
                pos,
                e1=self.nodes[pos][0],
                n1=self.nodes[pos][1],
                e2=self.nodes[pos + 1][0],
                n2=self.nodes[pos + 1][1],
            )
        if pos != 0:
            self._moveSegment(
                pos,
                e1=self.nodes[pos - 1][0],
                n1=self.nodes[pos - 1][1],
                e2=self.nodes[pos][0],
                n2=self.nodes[pos][1],
            )

    def __len__(self):
        return len(self.segments)

    def dislocSource(self):
        return np.array([seg.dislocSource() for seg in self.segments if seg.enabled])


class DislocProcessor(SourceProcessor):
    __implements__ = "disloc"

    @staticmethod
    def process(sources, sandbox, nthreads=0):
        result = {
            "processor_profile": dict(),
            "displacement.n": np.zeros(sandbox.frame.npixel),
            "displacement.e": np.zeros(sandbox.frame.npixel),
            "displacement.d": np.zeros(sandbox.frame.npixel),
        }

        src_nu = set(src.nu for src in sources)

        for nu in src_nu:
            nu_sources = [src for src in sources if src.nu == nu]
            nsources = len(nu_sources)
            src_arr = np.vstack([src.dislocSource() for src in nu_sources])

            north_shifts, east_shifts = od.latlon_to_ne_numpy(
                np.repeat(sandbox.frame.llLat, nsources),
                np.repeat(sandbox.frame.llLon, nsources),
                np.array([src.lat for src in nu_sources]),
                np.array([src.lon for src in nu_sources]),
            )

            src_arr[:, 5] += east_shifts
            src_arr[:, 6] += north_shifts

            res = disloc_ext.disloc(
                src_arr, sandbox.frame.coordinatesMeter, nu, nthreads
            )

            result["displacement.e"] += res[:, 0]
            result["displacement.n"] += res[:, 1]
            result["displacement.d"] += -res[:, 2]

        return result



import logging
import os
import re
import time

import requests
from requests.compat import urljoin

op = os.path

data_uri = "https://data.pyrocko.org/testing/kite/"
data_dir = op.join(op.dirname(op.abspath(__file__)), "data/")

logger = logging.getLogger("kite.testing")


class DownloadError(Exception):
    ...


def _makedir(path):
    try:
        os.makedirs(path)
    except OSError:
        ...


def get_test_data(fn):
    def _dir_content(url):
        r = requests.get(url)
        r.raise_for_status()

        entries = re.findall(r'href="([a-zA-Z0-9_.-]+/?)"', r.text)

        files = sorted(set(fn for fn in entries if not fn.endswith("/")))
        return files

    def _file_size(url):
        r = requests.head(url, headers={"Accept-Encoding": "identity"})
        r.raise_for_status()
        return int(r.headers["content-length"])

    def _download_file(url, fn_local):
        if op.exists(fn_local):
            if os.stat(fn_local).st_size == _file_size(url):
                logger.info("Using cached file %s" % fn_local)
                return fn_local

        logger.info("Downloading %s..." % url)
        fsize = _file_size(url)

        r = requests.get(url, stream=True)
        r.raise_for_status()

        dl_bytes = 0
        with open(fn_local, "wb") as f:
            for d in r.iter_content(chunk_size=1024):
                dl_bytes += len(d)
                f.write(d)

        if dl_bytes != fsize:
            raise DownloadError(
                f"Download {url} incomplete! Got {fsize} bytes, expected {dl_bytes}"
            )
        logger.info("Download completed.")
        return fn_local

    url = urljoin(data_uri, fn)

    dl_dir = data_dir
    _makedir(dl_dir)

    if fn.endswith("/"):
        dl_dir = op.join(data_dir, fn)
        _makedir(dl_dir)
        dl_files = _dir_content(url)

        dl_files = zip(
            [urljoin(url, u) for u in dl_files], [op.join(dl_dir, f) for f in dl_files]
        )
    else:
        dl_files = (url, op.join(data_dir, fn))
        return _download_file(*dl_files)

    return [_download_file(*f) for f in dl_files]


class Benchmark(object):
    def __init__(self, prefix=None):
        self.prefix = prefix or ""
        self.results = []

    def __call__(self, func):
        def stopwatch(*args):
            t0 = time.time()
            name = self.prefix + func.__name__
            result = func(*args)
            elapsed = time.time() - t0
            self.results.append((name, elapsed))
            return result

        return stopwatch

    def __str__(self):
        rstr = ["Benchmark results"]
        if self.prefix != "":
            rstr[-1] += " - %s" % self.prefix

        if len(self.results) > 0:
            indent = max([len(name) for name, _ in self.results])
        else:
            indent = 0
        rstr.append("=" * (indent + 17))
        rstr.insert(0, rstr[-1])
        for res in self.results:
            rstr.append("{0:<{indent}}{1:.8f} s".format(*res, indent=indent + 5))
        if len(self.results) == 0:
            rstr.append("None ran!")
        return "\n".join(rstr)


def setLogLevel(level):
    level = getattr(logging, level, "DEBUG")
    logging.basicConfig(level=level)



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.path import Path
from pyrocko.guts import Bool, Dict

from .plugin import Plugin, PluginConfig


class PolygonMaskConfig(PluginConfig):
    polygons = Dict.T(optional=True, default={})
    applied = Bool.T(default=True)


class PolygonMask(Plugin):
    def __init__(self, scene, config=None):
        self.scene = scene
        self.config = config or PolygonMaskConfig()
        self._points = None

        self._log = scene._log.getChild("Mask")

    @property
    def polygons(self):
        return self.config.polygons

    @property
    def npolygons(self):
        return len(self.config.polygons)

    def get_points(self):
        if self._points is None:
            f = self.scene.frame
            cols, rows = np.meshgrid(range(f.cols), range(f.rows))
            self._points = np.vstack((cols.ravel(), rows.ravel())).T
        return self._points

    def click_one_polygon(self):
        """
        Open a matplotlib window to click a closed polygon to mask.
        """

        sc = self.scene

        fig = plt.figure()
        ax = fig.add_axes([0.05, 0.05, 0.9, 0.9])
        ax.imshow(sc.displacement, origin="lower")
        ax.set_title("Click to add vertex. Press ENTER to finish.")

        #  Click polygon to mask
        vertices = plt.ginput(-1)
        self.add_polygon(vertices)

    def add_polygon(self, vertices):
        pid = len(self.config.polygons)
        while pid in self.config.polygons:
            pid += 1
        self.config.polygons[pid] = vertices
        self.update()

        return pid

    def remove_polygon(self, pid):
        self.config.polygons.pop(pid, None)
        self.update()

    def update_polygon(self, pid, vertices):
        if pid not in self.config.polygons:
            return
        self.config.polygons[pid] = vertices
        self.update()

    def apply(self, displacement):
        sc = self.scene
        points = self.get_points()

        mask = np.full((sc.frame.rows, sc.frame.cols), False)
        for vertices in self.config.polygons.values():
            p = Path(vertices)
            mask |= p.contains_points(points).reshape(sc.frame.rows, sc.frame.cols)

        displacement[mask] = np.nan
        return displacement



import shutil
import tempfile
import unittest

import numpy as np

from kite import SandboxScene, TestSandboxScene
from kite.sources import OkadaPath, OkadaSource

from . import common

plot = False
common.setLogLevel("DEBUG")


class testSourceOkada(unittest.TestCase):
    __name__ = "SandboxTestOkada"

    def setUp(self):
        self.ms = SandboxScene()
        self.tmpdir = tempfile.mkdtemp(prefix="kite")

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def test_okada_source(self):
        nsources = 2

        def random_int(low, high):
            return np.random.randint(low, high, size=1).astype(float)

        for s in range(nsources):
            length = random_int(5000, 15000)
            self.ms.addSource(
                OkadaSource(
                    easting=random_int(
                        self.ms.frame.E.min(), self.ms.frame.E.max()
                    ),  # ok
                    northing=random_int(
                        self.ms.frame.N.min(), self.ms.frame.N.max()
                    ),  # ok
                    depth=random_int(0, 8000),  # ok
                    strike=random_int(0, 360),  # ok
                    dip=random_int(0, 170),
                    slip=random_int(1, 5),  # ok
                    rake=random_int(0, 180),
                    length=length,
                    width=15.0 * length**0.66,
                )
            )

            self._plot_displacement(self.ms)

    def _test_okada_path(self):
        ok_path = OkadaPath(
            easting=10000,
            northing=24000,
        )
        ok_path.addNode(15000, 28000)
        ok_path.addNode(18000, 32000)
        ok_path.addNode(22000, 34000)
        # ok_path.insertNode(1, 22000, 34000)
        self.ms.addSource(ok_path)

        self._plot_displacement(self.ms)

    @staticmethod
    def _plot_displacement(ms):
        if not plot:
            ms.down
            return

        import matplotlib.pyplot as plt
        from matplotlib.patches import Polygon

        fig = plt.figure()
        ax = fig.gca()
        ms.processSources()

        ax.imshow(
            np.flipud(ms.down),
            aspect="equal",
            extent=[0, ms.frame.E.max(), 0, ms.frame.N.max()],
        )
        for src in ms.sources:
            for seg in src.segments:
                p = Polygon(seg.outline(), alpha=0.8, fill=False)
                ax.add_artist(p)
            if isinstance(src, OkadaPath):
                nodes = np.array(src.nodes)
                ax.scatter(nodes[:, 0], nodes[:, 1], color="r")
        plt.show()
        fig.clear()

    def testModelSaveLoad(self):
        filename = self.tmpdir + "/testsave.yml"
        msc = TestSandboxScene.randomOkada(nsources=2)
        msc.save(filename=filename)

        msd2 = SandboxScene.load(filename=filename)  # noqa
        # print msc2.config


if __name__ == "__main__":
    plot = True
    unittest.main(exit=False)



import matplotlib.animation as animation
import matplotlib.pyplot as plt
import numpy as num

from kite import SandboxScene
from kite.sources import OkadaSource

km = 1e3

sandbox = SandboxScene()

nframes = 20
depths = num.linspace(4 * km, 4 * km, nframes)
strikes = num.linspace(0, 180, nframes)

okada = OkadaSource(
    northing=40 * km,
    easting=40 * km,
    depth=4 * km,
    length=8 * km,
    width=4 * km,
    strike=63.0,
    rake=0,
    dip=0.0,
    slip=3.0,
    opening=0,
)

sandbox.addSource(okada)

sandbox.processSources()

fig, axis = plt.subplots(nrows=2, ncols=2)


# Plot the resulting surface displacements


def imargs(data):
    max_value = max(num.abs(data.max()), num.abs(data.min()))
    return {
        "X": data,
        "cmap": "bwr",
        "vmin": -max_value,
        "vmax": max_value,
        "animated": True,
    }


components = [sandbox.north, sandbox.east, sandbox.down, sandbox.displacement]
titles = ["North", "East", "Down", "LOS"]
images = []


for ax, comp, title in zip(fig.axes, components, titles):
    im = ax.imshow(**imargs(comp))
    images.append(im)
    ax.set_title(title)


def update_figure(iframe, *args):
    print("Updating figure! (frame %03d)" % iframe)

    okada.depth = depths[iframe]
    okada.strike = strikes[iframe]
    sandbox.processSources()

    for im, comp in zip(images, components):
        args = imargs(comp)
        im.set_data(comp)
    return images


ani = animation.FuncAnimation(
    fig, update_figure, interval=50, frames=nframes, blit=True
)

fig.show()



import glob
import logging
from datetime import datetime, timedelta
from os import path as op

import numpy as np
import pyrocko.orthodrome as od
from scipy import interpolate, io, stats

try:
    import h5py
except ImportError as e:
    raise e("Please install h5py library")

import matplotlib.pyplot as plt
from matplotlib import colors

from kite.scene import Scene, SceneConfig

log = logging.getLogger("stamps2kite")

d2r = np.pi / 180.0
r2d = 180.0 / np.pi


class DataStruct(dict):
    def __getattr__(self, attr):
        return self[attr]

    def __setattr__(self, attr, value):
        self[attr] = value


def _get_file(dirname, fname):
    fns = glob.glob(op.join(dirname, fname))
    if len(fns) == 0:
        raise ImportError("Cannot find %s file in %s" % (fname, dirname))
    if len(fns) > 1:
        raise ImportError("Found multiple files for %s: %s" % (fname, ", ".join(fns)))
    fn = fns[0]

    log.debug("Found file %s", fn)
    return fn


def _read_mat(filename):
    try:
        mat = h5py.File(filename, "r")
    except OSError:
        log.debug("using old scipy import for %s", filename)
        mat = io.loadmat(filename)
    return mat


def read_mat_data(dirname, import_mv2=False, **kwargs):
    # TODO: Add old matlab import
    log.debug("Reading in StaMPS project...")

    fn_ps2 = kwargs.get("fn_ps2", _get_file(dirname, "ps2.mat"))
    fn_ps_plot = kwargs.get("fn_ps_plot", _get_file(dirname, "ps_plot*.mat"))
    fn_parms = kwargs.get("fn_parms", _get_file(dirname, "parms.mat"))
    fn_width = kwargs.get("fn_width", _get_file(dirname, "width.txt"))
    fn_len = kwargs.get("fn_len", _get_file(dirname, "len.txt"))
    fn_look_angle = kwargs.get("fn_look_angle", _get_file(dirname, "look_angle.1.in"))

    ps2_mat = _read_mat(fn_ps2)
    ps_plot_mat = _read_mat(fn_ps_plot)
    params_mat = _read_mat(fn_parms)

    data = DataStruct()
    data.ll_coords = np.asarray(ps2_mat["ll0"])
    data.radar_coords = np.asarray(ps2_mat["ij"])
    data.ps_mean_v = np.asarray(ps_plot_mat["ph_disp"]).ravel()

    geo_coords = np.asarray(ps2_mat["lonlat"])
    data.lons = geo_coords[0, :]
    data.lats = geo_coords[1, :]

    days = np.asarray(ps2_mat["day"])
    data.tmin = timedelta(days=days.min() - 366.25) + datetime(1, 1, 1)
    data.tmax = timedelta(days=days.max() - 366.25) + datetime(1, 1, 1)

    if import_mv2:
        fn_mv2 = kwargs.get("fn_mv2", _get_file(dirname, "mv2.mat"))
        mv2_mat = h5py.File(fn_mv2, "r")
        data.ps_mean_std = np.asarray(mv2_mat["mean_v_std"]).ravel()

    with open(fn_len) as rl, open(fn_width) as rw:
        data.px_length = int(rl.readline())
        data.px_width = int(rw.readline())

    data.look_angles = np.loadtxt(fn_look_angle)[::2]

    heading = float(np.asarray(params_mat["heading"]))
    if np.isnan(heading):
        raise ValueError("Heading information in parms.mat is missing!")

    data.heading = heading

    return data


def bin_ps_data(data, bins=(800, 800)):
    log.debug("Binning StaMPS velocity data...")
    bin_vels, edg_lat, edg_lon, _ = stats.binned_statistic_2d(
        data.lats, data.lons, data.ps_mean_v, statistic="mean", bins=bins
    )

    log.debug("Binning radar coordinates...")
    bin_radar_i, _, _, _ = stats.binned_statistic_2d(
        data.lats, data.lons, data.radar_coords[1], statistic="mean", bins=bins
    )
    bin_radar_j, _, _, _ = stats.binned_statistic_2d(
        data.lats, data.lons, data.radar_coords[2], statistic="mean", bins=bins
    )

    if "ps_mean_std" in data.keys():
        log.debug("Binning mean velocity variance...")
        bin_mean_std, _, _, _ = stats.binned_statistic_2d(
            data.lats, data.lons, data.ps_mean_std, statistic="mean", bins=bins
        )
        data.bin_mean_var = bin_mean_std**2  # We want variance

    data.bin_ps_mean_v = bin_vels

    data.bin_radar_i = bin_radar_i
    data.bin_radar_j = bin_radar_j

    data.bin_edg_lat = edg_lat
    data.bin_edg_lon = edg_lon

    return data


def interpolate_look_angles(data):
    log.info("Interpolating look angles from radar coordinates...")
    log.debug(
        "Radar coordinates extent width %d; length %d", data.px_width, data.px_length
    )
    log.debug(
        "Radar coordinates data: length %d - %d; width %d - %d",
        data.radar_coords[1].min(),
        data.radar_coords[1].max(),
        data.radar_coords[0].min(),
        data.radar_coords[0].max(),
    )
    log.debug(
        "Binned radar coordinate ranges: length %d - %d; width %d - %d",
        np.nanmin(data.bin_radar_i),
        np.nanmax(data.bin_radar_i),
        np.nanmin(data.bin_radar_j),
        np.nanmax(data.bin_radar_j),
    )

    width_coords = np.linspace(0, data.px_width, 50)
    len_coords = np.linspace(0, data.px_length, 50)
    coords = np.asarray(np.meshgrid(width_coords, len_coords)).reshape(2, 2500)

    radar_coords = np.vstack(
        [
            data.bin_radar_j.ravel() - data.radar_coords[0].min(),
            data.bin_radar_i.ravel() - data.radar_coords[1].min(),
        ]
    )

    interp = interpolate.LinearNDInterpolator(coords.T, data.look_angles)
    data.bin_look_angles = interp(radar_coords.T).reshape(*data.bin_ps_mean_v.shape)
    return interp


def stamps2kite(
    dirname=".", px_size=(800, 800), convert_m=True, import_var=False, **kwargs
):
    """Convert StaMPS velocity data to a Kite Scene

    Loads the mean PS velocities (from e.g. ``ps_plot(..., -1)``) from a
    StaMPS project, and grids the data into mean velocity bins. The LOS
    velocities will be converted to a Kite Scene (:class:`~kite.Scene`).

    If explicit filenames `fn_ps` or `fn_ps_plot` are not passed an
    auto-discovery of :file:`ps?.mat` and :file:`ps_plot*.mat` in
    directory :param:`dirname` is approached.

    .. note ::

        Running the stamps2kite script on your StaMPS project does
        the trick in most cases.

    :param dirname: Folder containing data from the StaMPS project.
        Defaults to ``.``.
    :type dirname: str
    :param fn_ps2: :file:`ps2.mat` file with lat/lon information about the PS
        scene. Defaults to ``ps2.mat``.
    :type fn_ps2: str
    :param fn_mv2: :file:`mv2.mat` file with mean velocity standard deviations
        of every PS point, created by `ps_plot(...)`. Defaults to ``mv2.mat``.
    :type fn_mv2: str
    :param fn_ps_plot: Processed output from StaMPS ``ps_plot`` function, e.g.
        :file:`ps_plot*.mat`. Defaults to ``ps_plot*.mat``.
    :type fn_ps_plot: str
    :param fn_parms: :file:`parms.mat` from StaMPS, holding essential meta-
        information. Defaults to ``parms.mat``.
    :type fn_parms: str
    :param fn_look_angle: The :file:`look_angle.1.in` holds the a 50x50 grid
        with look angle information. Defaults to ``look_angle.1.in``.
    :type fn_look_angle: str
    :param fn_width: The :file:`width.txt` containing number of radar columns.
        Defaults to ``width.txt``.
    :type fn_width: str
    :param fn_len: The :file:`len.txt` containing number of rows in the
        interferogram. Defaults to ``len.txt``.
    :type fn_len: str
    :param px_size: Size of pixels in North and East in meters.
        Default (200, 200).
    :type px_size: tuple
    :param convert_m: Convert displacement to meters, default True.
    :type convert_m: bool
    :param import_var: Import the mean velocity variance, this information
        is used by the Kite scene to define the covariance.
    :param import_var: bool

    :returns: Kite Scene from the StaMPS data
    :rtype: :class:`kite.Scene`
    """
    data = read_mat_data(dirname, import_mv2=import_var, **kwargs)
    log.info("Found a StaMPS project at %s", op.abspath(dirname))

    bbox = (data.lons.min(), data.lats.min(), data.lons.max(), data.lats.max())

    lengthN = od.distance_accurate50m(bbox[1], bbox[0], bbox[3], bbox[0])
    lengthE = od.distance_accurate50m(bbox[1], bbox[0], bbox[1], bbox[2])
    bins = (round(lengthE / px_size[0]), round(lengthN / px_size[1]))

    if convert_m:
        data.ps_mean_v /= 1e3

    if convert_m and import_var:
        data.ps_mean_std /= 1e3

    bin_ps_data(data, bins=bins)
    interpolate_look_angles(data)

    log.debug("Processing of LOS angles")
    data.bin_theta = data.bin_look_angles * d2r

    phi_angle = -data.heading * d2r + np.pi
    if phi_angle > np.pi:
        phi_angle -= 2 * np.pi
    data.bin_phi = np.full_like(data.bin_theta, phi_angle)
    data.bin_phi[np.isnan(data.bin_theta)] = np.nan

    log.debug("Setting up the Kite Scene")
    config = SceneConfig()
    config.frame.llLat = data.bin_edg_lat.min()
    config.frame.llLon = data.bin_edg_lon.min()
    config.frame.dE = data.bin_edg_lon[1] - data.bin_edg_lon[0]
    config.frame.dN = data.bin_edg_lat[1] - data.bin_edg_lat[0]
    config.frame.spacing = "degree"

    scene_name = op.basename(op.abspath(dirname))
    config.meta.scene_title = "%s (StamPS import)" % scene_name
    config.meta.scene_id = scene_name
    config.meta.time_master = data.tmin.timestamp()
    config.meta.time_slave = data.tmax.timestamp()

    scene = Scene(
        theta=data.bin_theta,
        phi=data.bin_phi,
        displacement=data.bin_ps_mean_v,
        config=config,
    )

    if import_var:
        scene.displacement_px_var = data.bin_mean_var

    return scene


def plot_scatter_vel(lat, lon, vel):
    vmax = abs(max(vel.min(), vel.max()))
    norm = colors.Normalize(vmin=-vmax, vmax=vmax)

    plt.scatter(lat, lon, s=5, c=vel, cmap="RdYlBu", norm=norm)
    plt.show()


def main():
    import argparse

    parser = argparse.ArgumentParser(
        description="""Convert StaMPS data into a Kite scene.

Loads the PS velocities from a StaMPS project (i.e. from ps_plot(..., -1);),
and grids the data into mean velocity bins. The mean LOS velocities will be
converted into a Kite Scene.

The data has to be fully processed through StaMPS and may stem from the master
project or the processed small baseline pairs. Required files are:

 - ps2.mat          Meta information and geographical coordinates.
 - parms.mat        Meta information about the scene (heading, etc.).
 - ps_plot*.mat     Processed mean LOS velocities from ps_plot(..., -1).
 - mv2.mat          Mean velocity's standard deviation from ps_plot(..., -1).

 - look_angle.1.in  Look angles for the scene.
 - width.txt        Width dimensions of the interferogram and
 - len.txt          length.
 """,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument("folder", type=str, default=".", help="StaMPS project folder.")
    parser.add_argument(
        "--resolution",
        "-r",
        nargs=2,
        metavar=("mN", "mE"),
        dest="resolution",
        type=int,
        default=(500, 500),
        help="pixel size of the output grid in North and East (meter)."
        "Default is 500 m by 500 m.",
    )
    parser.add_argument(
        "--save",
        "-s",
        default=None,
        type=str,
        dest="save",
        help="filename to save the Kite scene to. If not given, the scene"
        " will be opened in spool GUI.",
    )
    parser.add_argument(
        "--force",
        "-f",
        default=False,
        action="store_true",
        dest="force",
        help="force overwrite of an existing scene.",
    )
    parser.add_argument(
        "--keep-mm",
        action="store_true",
        default=False,
        help="keep mm/a and do not convert to m/a.",
    )
    parser.add_argument(
        "--import-var",
        action="store_true",
        dest="import_var",
        default=False,
        help="import the variance from mv2.mat, which is added to Kite's"
        " scene covariance matrix.",
    )
    parser.add_argument(
        "-v",
        action="count",
        default=0,
        help="verbosity, add multiple to increase verbosity.",
    )

    args = parser.parse_args()

    log_level = logging.INFO - args.v * 10
    logging.basicConfig(level=log_level if log_level > 0 else 0)

    fn_save = args.save
    if args.save:
        for fn in (fn_save, fn_save + ".yml", fn_save + ".npz"):
            if op.exists(fn) and not args.force:
                raise UserWarning("File %s exists! Use --force to overwrite." % fn_save)

    scene = stamps2kite(
        dirname=args.folder,
        px_size=args.resolution,
        convert_m=not args.keep_mm,
        import_var=args.import_var,
    )

    if fn_save:
        fn_save.rstrip(".yml")
        fn_save.rstrip(".npz")

        log.info("Saving StaMPS scene to file %s[.yml/.npz]...", fn_save)
        scene.save(args.save)

    else:
        scene.spool()


if __name__ == "__main__":
    main()



#!/bin/python
import argparse as ap
import logging
import sys

from kite import Scene, TestScene
from kite.spool import spool


def main(args=None):
    """
    Spool app deployed through setuptools
    """
    if args is None:
        args = sys.argv[1:]

    epilog = """Spool is part of the kite InSAR framework.

Author: Marius Paul Isken (marius.isken@gfz-potsdam.de)
Documentation: https://pyrocko.org"""
    desc = "InSAR deformation inspector, quadtree and covariance"

    parser = ap.ArgumentParser(
        prog="spool",
        epilog=epilog,
        description=desc,
        parents=[],
        formatter_class=ap.RawTextHelpFormatter,
        prefix_chars="-",
        fromfile_prefix_chars=None,
        argument_default=ap.SUPPRESS,
        conflict_handler="resolve",
        add_help=True,
    )
    parser.add_argument(
        "file",
        type=str,
        help="Load native kite container (.npz & .yml)",
        default=None,
        nargs="?",
    )
    parser.add_argument(
        "--load",
        metavar="file",
        type=str,
        default=None,
        help="""Import file or directory
Supported formats are:
 - Matlab   *.mat
 - GAMMA    * binary and *.par file
 - GMTSAR   *.grd binary and *.los.* file
 - ISCE     *.unw.geo with *.unw.geo.xml and; *.rdr.geo for LOS data
 - ROI_PAC  * binary and *.rsc file
 - SARSCAPE *los_ll.grd and *.los.enu file
 - SNAP     *.rsc and *.Abstracted_Metadata.txt file (GAMMA Export)
 - ARIA     Extracted layers: unwrappedPhase, lookAngle, incidenceAngle,
             connectedComponents
 - LiCSAR   *.unw.tif and LOS data, see client.download_licsar

 For more information see the online documentation at https://pyrocko.org""",
    )
    parser.add_argument(
        "--synthetic",
        type=str,
        default=None,
        choices=["fractal", "sine", "gauss"],
        help="""Synthetic Tests
Available Synthetic Displacement:
 * fractal (Atmospheric model; after Hanssen, 2001)
 * sine
 * gauss
""",
    )

    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=1,
        help="Verbosity, add multiple to increase verbosity.",
    )

    ns = parser.parse_args(args)

    log_level = logging.WARNING - ns.verbose * 10
    log_level = log_level if log_level > logging.DEBUG else logging.DEBUG
    logging.basicConfig(level=log_level if log_level > 0 else 0)

    if ns.load is None and ns.synthetic is None and ns.file is None:
        parser.print_help()
        sys.exit(0)

    sc = None
    if ns.synthetic is not None:
        if ns.synthetic == "fractal":
            sc = TestScene.createFractal()
        elif ns.synthetic == "sine":
            sc = TestScene.createSine()
        elif ns.synthetic == "gauss":
            sc = TestScene.createFractal()
        else:
            parser.print_help()
            sys.exit(0)

    elif ns.file is not None:
        sc = Scene.load(ns.file)

    if sc:
        spool(scene=sc)
    elif ns.load is not None:
        spool(import_file=ns.load)


if __name__ == "__main__":
    main()



import numpy as np
import pyqtgraph as pg
from PyQt5 import QtCore, QtGui, QtWidgets
from pyqtgraph import dockarea

from .config import getConfig
from .util import SourceROI

d2r = np.pi / 180.0


class SandboxSceneLayout(pg.GraphicsLayoutWidget):

    PLOT_VIEWS = ["north", "east", "down", "los"]

    def __init__(self, sandbox, *args, **kwargs):
        pg.GraphicsLayoutWidget.__init__(self, **kwargs)
        self.sandbox = sandbox

        self.plots = [
            DisplacementPlot(sandbox, title="North", component=lambda m: m.north),
            DisplacementPlot(sandbox, title="East", component=lambda m: m.east),
            DisplacementVectorPlot(sandbox, title="Down", component=lambda m: m.down),
            DisplacementPlot(sandbox, title="LOS", component=lambda m: m.displacement),
        ]

        for plt in self.plots:
            plt.vb.menu = QtWidgets.QMenu(self)

        self.updateViews()
        getConfig().qconfig.updated.connect(self.updateViews)

        self._mov_sig = pg.SignalProxy(
            self.scene().sigMouseMoved, rateLimit=60, slot=self.mouseMoved
        )

    @QtCore.pyqtSlot()
    def updateViews(self):
        self.clear()
        for plt in self.plots:
            plt.removeHintText()

        config_mask = [
            getConfig().__getattribute__(cfg)
            for cfg in ("view_north", "view_east", "view_down", "view_los")
        ]

        visible_plots = [plt for ip, plt in enumerate(self.plots) if config_mask[ip]]

        for ip, plt in enumerate(visible_plots):
            row = ip / 2
            col = ip % 2 + 1

            self.addItem(plt, row=row, col=col)
            plt.showGrid(x=True, y=True)
            plt.hideAxis("bottom")
            plt.hideAxis("left")
            plt.vb.border = pg.mkPen(50, 50, 50)
            if ip != 0:
                plt.setXLink(self.plots[0])
                plt.setYLink(self.plots[0])

        if len(visible_plots) > 0:
            visible_plots[-1].addHintText()
            visible_plots[-1].autoRange(items=[visible_plots[-1].image])

    def resizeEvent(self, ev):
        pg.GraphicsLayoutWidget.resizeEvent(self, ev)
        if hasattr(self, "plots"):
            viewbox = self.plots[0].getViewBox()
            viewbox.autoRange()

    @QtCore.pyqtSlot(object)
    def mouseMoved(self, event):
        self.sandbox.cursor_tracker.sigMouseMoved.emit(event)


class ModelReferenceLayout(pg.GraphicsLayoutWidget):
    def __init__(self, sandbox, *args, **kwargs):
        pg.GraphicsLayoutWidget.__init__(self, **kwargs)
        self.sandbox = sandbox

        self.plots = [
            DisplacementPlot(
                sandbox,
                title="Scene Displacement",
                component=lambda m: m.reference.scene.displacement,
            ),
            DisplacementPlot(
                sandbox,
                title="Model Residual",
                component=lambda m: m.reference.difference,
            ),
        ]
        self.plots[-1].addHintText()

        self._mov_sig = pg.SignalProxy(
            self.scene().sigMouseMoved, rateLimit=60, slot=self.mouseMoved
        )

        for ip, plt in enumerate(self.plots):
            row = ip / 2
            col = ip % 2 + 1

            self.addItem(plt, row=row, col=col)
            plt.showGrid(x=True, y=True)
            plt.hideAxis("bottom")
            plt.hideAxis("left")
            plt.vb.border = pg.mkPen(50, 50, 50)
            if ip != 0:
                plt.setXLink(self.plots[0])
                plt.setYLink(self.plots[0])

        def getAxis(plt, orientation, label):
            axis = pg.AxisItem(orientation=orientation, linkView=plt.vb)
            axis.setLabel(label, units="m")
            return axis

        plts = self.plots
        self.addItem(getAxis(plts[0], "left", "Northing"), row=0, col=0)
        self.addItem(getAxis(plts[1], "left", "Northing"), row=1, col=0)
        self.addItem(getAxis(plts[0], "bottom", "Easting"), row=2, col=1)
        self.addItem(getAxis(plts[1], "bottom", "Easting"), row=2, col=2)

    @QtCore.pyqtSlot(object)
    def mouseMoved(self, event):
        self.sandbox.cursor_tracker.sigMouseMoved.emit(event)


class CursorRect(QtWidgets.QGraphicsRectItem):
    pen = pg.mkPen((0, 0, 0, 120), width=1.0)
    cursor = QtCore.QRectF((QtCore.QPointF(-1.5, -1.5)), (QtCore.QPointF(1.5, 1.5)))

    def __init__(self):
        QtWidgets.QGraphicsRectItem.__init__(self, self.cursor)
        self.setPen(self.pen)
        self.setZValue(1e9)
        self.setFlag(QtWidgets.QGraphicsItem.ItemIgnoresTransformations)


class DisplacementPlot(pg.PlotItem):
    def __init__(self, sandbox, component, title="Untitled"):
        pg.PlotItem.__init__(self)
        self.title = title
        self.sandbox = sandbox
        self.component = component

        self.cursor = CursorRect()
        self.addCursor()

        self.setAspectLocked(True)
        self.setLabels(bottom=("Easting", "m"), left=("Northing", "m"))

        border_pen = pg.mkPen(255, 255, 255, 50)

        self.image = pg.ImageItem(
            None, autoDownsample=False, border_pen=border_pen, useOpenGL=True
        )
        self.addItem(self.image)

        self.title_label = pg.LabelItem(
            text='<span style="color: #9E9E9E;">' "%s</span>" % self.title,
            justify="right",
            size="10pt",
            parent=self,
        )
        self.title_label.anchor(itemPos=(0.0, 0.0), parentPos=(0.01, 0.01))
        self.title_label.setOpacity(0.6)

        self.hint_text = None

        self.sandbox.sigModelUpdated.connect(self.update)
        self.sandbox.sources.modelAboutToBeReset.connect(self.removeSourceROIS)
        self.sandbox.sources.modelReset.connect(self.addSourceROIS)

        self.update()
        self.rois = []
        self.addSourceROIS()

    @property
    def data(self):
        return self.component(self.sandbox.model)

    @QtCore.pyqtSlot()
    def update(self):
        self.image.updateImage(self.data.T)
        self.transFromFrame()

    def transFromFrame(self):
        self.image.resetTransform()
        self.image.scale(self.sandbox.frame.dEmeter, self.sandbox.frame.dNmeter)

    @QtCore.pyqtSlot()
    def addSourceROIS(self):
        self.rois = []
        index = QtCore.QModelIndex()
        for isrc in range(self.sandbox.sources.rowCount(index)):
            index = self.sandbox.sources.index(isrc, 0, index)
            roi = index.data(SourceROI)
            roi.setParent(self)
            self.rois.append(roi)
            self.addItem(roi)

    @QtCore.pyqtSlot()
    def removeSourceROIS(self):
        if self.rois:
            for roi in self.rois:
                self.removeItem(roi)
            self.update()

    def addCursor(self):
        self.sandbox.cursor_tracker.sigCursorMoved.connect(self.drawCursor)
        self.sandbox.cursor_tracker.sigMouseMoved.connect(self.mouseMoved)
        self.addItem(self.cursor)

    @QtCore.pyqtSlot(object)
    def mouseMoved(self, event):

        if self.vb.sceneBoundingRect().contains(event[0]):
            map_pos = self.vb.mapSceneToView(event[0])

            img_pos = self.image.mapFromScene(event[0])
            pE, pN = img_pos.x(), img_pos.y()
            if (pE < 0 or pN < 0) or (
                pE > self.image.image.shape[0] or pN > self.image.image.shape[1]
            ):
                value = 0.0
            else:
                value = self.image.image[int(img_pos.x()), int(img_pos.y())]
            self.sandbox.cursor_tracker.sigCursorMoved.emit((map_pos, value))
            self.cursor.hide()
        else:
            if not getConfig().show_cursor:
                self.cursor.hide()
            else:
                self.cursor.show()

    @QtCore.pyqtSlot(object)
    def drawCursor(self, pos):
        pos, _ = pos
        self.cursor.setPos(pos)

    def addHintText(self):
        self.hint_text = pg.LabelItem(text="", justify="right", size="8pt", parent=self)
        self.hint_text.anchor(itemPos=(1.0, 1.0), parentPos=(1.0, 1.0))
        self.hint_text.text_template = (
            '<span style="font-family: monospace; color: #fff;'
            'background-color: #000;">'
            "East {0:08.2f} m | North {1:08.2f} m | "
            "Displacement {2:2.4f} m</span>"
        )
        self.hint_text.setOpacity(0.6)
        self.sandbox.cursor_tracker.sigCursorMoved.connect(self.updateHintText)

    def removeHintText(self):
        if self.hint_text is not None:
            self.removeItem(self.hint_text)
            self.hint_text = None

    @QtCore.pyqtSlot(object)
    def updateHintText(self, pos):
        pos, value = pos
        if self.hint_text is not None:
            self.hint_text.setText(
                self.hint_text.text_template.format(pos.x(), pos.y(), value)
            )


class DisplacementVectorPlot(DisplacementPlot):
    def __init__(self, *args, **kwargs):
        DisplacementPlot.__init__(self, *args, **kwargs)

        vectors = DisplacementVectors(self)
        self.addItem(vectors)


class DisplacementVectors(QtWidgets.QGraphicsItemGroup):
    def __init__(self, plot, *args, **kwargs):
        QtWidgets.QGraphicsItemGroup.__init__(self, parent=plot, *args, **kwargs)
        self.vb = plot.vb
        self.plot = plot
        self.image = plot.image
        self.sandbox = plot.sandbox

        self.scale_view = None
        self.scale_length = None
        self.vectors = []

        self.vb.geometryChanged.connect(self.prepareGeometryChange)
        getConfig().qconfig.updated.connect(self.createVectors)
        getConfig().qconfig.updated.connect(self.updateVectorAppearance)

        self.createVectors()

    def createVectors(self):
        for vec in self.vectors:
            vec.hide()

        if len(self.vectors) < getConfig().nvectors:
            while len(self.vectors) < getConfig().nvectors:
                vec = Vector(self)
                self.vectors.append(vec)
                self.addToGroup(vec)

    def updateVectorAppearance(self):
        Vector.arrow_color.setRgb(*getConfig().vector_color)
        Vector.arrow_brush.setColor(Vector.arrow_color)
        Vector.arrow_pen.setBrush(Vector.arrow_brush)
        Vector.arrow_pen.setWidth(getConfig().vector_pen_thickness)
        Vector.relative_length = getConfig().vector_relative_length

    def boundingRect(self):
        return self.vb.viewRect()

    def paint(self, painter, option, parent):
        r = self.vb.viewRect()
        h = r.height()
        w = r.width()

        nvectors = getConfig().nvectors

        nx = int(np.sqrt(nvectors) * float(w) / h)
        ny = int(np.sqrt(nvectors) * float(h) / w)
        dx = float(w) / nx
        dy = float(h) / ny
        d = dx if dx < dy else dy

        mat_N = self.sandbox.model.north.T
        mat_E = self.sandbox.model.east.T
        img_shape = self.image.image.shape
        ivec = 0

        length_scale = self.sandbox.model.max_horizontal_displacement
        self.length_scale = length_scale if length_scale > 0.0 else 1.0
        self.scale_view = (w + h) / 2 / painter.window().height() * 2.5

        for ix in range(nx):
            for iy in range(ny):
                if ivec > nvectors:
                    break
                vec = self.vectors[ivec]
                pos = QtCore.QPointF(r.x() + ix * dx + dx / 2, r.y() + iy * dy + dy / 2)

                # Slowest operation
                img_pos = self.plot.image.mapFromScene(self.vb.mapViewToScene(pos))

                pE = int(img_pos.x())
                pN = int(img_pos.y())

                if (
                    (pE >= img_shape[0] or pN >= img_shape[1])
                    or (pE < 0 or pN < 0)
                    or np.isnan(mat_E[pE, pN])
                    or np.isnan(mat_N[pE, pN])
                ):
                    dE = 0.0
                    dN = 0.0
                else:
                    dE = mat_E[pE, pN]
                    dN = mat_N[pE, pN]
                    dE = dE / self.length_scale * (d / self.scale_view)
                    dN = dN / self.length_scale * (d / self.scale_view)
                vec.setPos(pos)
                vec.setOrientation(dE, dN)

                if vec.scale() != self.scale_view:
                    vec.setScale(self.scale_view)
                vec.setVisible(True)

                ivec += 1

        while ivec < nvectors:
            self.vectors[ivec].hide()
            ivec += 1

        QtWidgets.QGraphicsItemGroup.paint(self, painter, option, parent)


class Vector(QtWidgets.QGraphicsItem):

    arrow_color = QtGui.QColor(*getConfig().vector_color)
    arrow_brush = QtGui.QBrush(arrow_color, QtCore.Qt.SolidPattern)
    arrow_pen = QtGui.QPen(
        arrow_brush, 1, QtCore.Qt.SolidLine, QtCore.Qt.RoundCap, QtCore.Qt.RoundJoin
    )

    relative_length = getConfig().vector_relative_length

    def __init__(self, parent):
        QtWidgets.QGraphicsItem.__init__(self, parent=parent)

        self.p1 = QtCore.QPointF()
        self.p2 = QtCore.QPointF()
        self.line = QtCore.QLineF(self.p1, self.p2)

        self.setOrientation(0.0, 0.0)
        self.setZValue(10000)

    def boundingRect(self):
        return QtCore.QRectF(self.p1, self.p2)

    def setOrientation(self, dEast, dNorth):
        dEast *= self.relative_length / 100
        dNorth *= self.relative_length / 100
        self.p2.setX(dEast)
        self.p2.setY(dNorth)
        self.line.setP2(self.p2)

    def paint(self, painter, option, parent):
        if self.line.length() == 0.0:
            return
        painter.setPen(self.arrow_pen)
        painter.setRenderHint(QtGui.QPainter.Antialiasing)
        painter.setCompositionMode(QtGui.QPainter.CompositionMode_SourceOver)

        arrow_length = self.line.length() * 0.3 * (self.relative_length / 100.0)
        d = self.line.angle()
        head_p1 = self.p2 - QtCore.QPointF(
            np.sin(d * d2r + np.pi / 3) * arrow_length,
            np.cos(d * d2r + np.pi / 3) * arrow_length,
        )

        head_p2 = self.p2 - QtCore.QPointF(
            np.sin(d * d2r + np.pi - np.pi / 3) * arrow_length,
            np.cos(d * d2r + np.pi - np.pi / 3) * arrow_length,
        )

        painter.drawLine(self.line)
        painter.drawPolyline(*[head_p1, self.p2, head_p2])


class ColormapPlots(pg.HistogramLUTWidget):
    def __init__(self):
        pg.HistogramLUTWidget.__init__(self, image=None)
        self.plots = []

        self.axis.setLabel("Displacement / m")

        zero_marker = pg.InfiniteLine(pos=0, angle=0, pen="w", movable=False)
        zero_marker.setValue(0.0)
        zero_marker.setZValue(1000)
        self.vb.addItem(zero_marker)

        self.axis.setLabel("Displacement / m")
        self.setSymColormap()

    @QtCore.pyqtSlot()
    def setSymColormap(self):
        cmap = {
            "ticks": [
                [0, (0, 0, 0, 255)],
                [1e-3, (106, 0, 31, 255)],
                [0.5, (255, 255, 255, 255)],
                [1.0, (8, 54, 104, 255)],
            ],
            "mode": "rgb",
        }
        cmap = {
            "ticks": [
                [0, (0, 0, 0)],
                [1e-3, (172, 56, 56)],
                [0.5, (255, 255, 255)],
                [1.0, (51, 53, 120)],
            ],
            "mode": "rgb",
        }

        lvl_min = lvl_max = 0.0
        for plot in self.plots:
            plt_min = np.nanmin(plot.data)
            plt_max = np.nanmax(plot.data)
            lvl_max = lvl_max if plt_max < lvl_max else plt_max
            lvl_min = lvl_min if plt_min > lvl_min else plt_min

        abs_range = max(abs(lvl_min), abs(lvl_max)) * 1.01

        self.gradient.restoreState(cmap)
        self.setLevels(-abs_range, abs_range)

    def addPlot(self, plot):
        image = plot.image
        if not self.plots:
            self.setImageItem(image)

        self.plots.append(plot)
        self.setSymColormap()

        # hist_pen = pg.mkPen((170, 57, 57, 255), width=1.)
        image.setLookupTable(self.getLookupTable)

        def updateLevels():
            image.setLevels(self.region.getRegion())

        self.sigLevelChangeFinished.connect(updateLevels)
        self.sigLevelsChanged.connect(updateLevels)
        updateLevels()


class SandboxSceneDockarea(dockarea.DockArea):
    def __init__(self, sandbox, *args, **kwargs):
        dockarea.DockArea.__init__(self)
        layout = SandboxSceneLayout(sandbox)

        cmap = ColormapPlots()
        for plt in layout.plots:
            cmap.addPlot(plt)
        sandbox.sigModelUpdated.connect(cmap.setSymColormap)

        cmap_dock = dockarea.Dock("Colormap", widget=cmap)
        cmap_dock.setStretch(1, None)

        layout_dock = dockarea.Dock("Model Sandbox", widget=layout)
        self.addDock(layout_dock, position="right")
        self.addDock(cmap_dock, position="right")


class ModelReferenceDockarea(dockarea.DockArea):
    def __init__(self, sandbox, *args, **kwargs):
        dockarea.DockArea.__init__(self)
        layout = ModelReferenceLayout(sandbox)

        cmap = ColormapPlots()
        for plt in layout.plots:
            cmap.addPlot(plt)

        cmap_dock = dockarea.Dock("Colormap", widget=cmap)
        cmap_dock.setStretch(1, None)

        layout_dock = dockarea.Dock("Model Sandbox", widget=layout)
        self.addDock(layout_dock, position="right")
        self.addDock(cmap_dock, position="right")



import numpy as np

from kite.sources import OkadaSource

from .base import RectangularSourceROI, SourceDelegate, SourceEditDialog

d2r = np.pi / 180.0
r2d = 180.0 / np.pi


class OkadaSourceDelegate(SourceDelegate):

    __represents__ = "OkadaSource"

    display_backend = "Okada"
    display_name = "OkadaSource"

    parameters = [
        "easting",
        "northing",
        "width",
        "length",
        "depth",
        "slip",
        "opening",
        "strike",
        "dip",
        "rake",
        "nu",
    ]
    ro_parameters = ["seismic_moment", "moment_magnitude"]

    class OkadaDialog(SourceEditDialog):
        def __init__(self, delegate, *args, **kwargs):
            super().__init__(delegate, ui_file="okada_source.ui", *args, **kwargs)

            def setLabel(method, fmt, value, suffix=""):
                method(fmt.format(value) + suffix)

            self.moment_magnitude.setValue = lambda v: setLabel(
                self.moment_magnitude.setText, "{:.2f}", v
            )
            self.seismic_moment.setValue = lambda v: setLabel(
                self.seismic_moment.setText, "{:.2e}", v, " Nm"
            )

            self.getSourceParameters()

    ROIWidget = RectangularSourceROI
    EditDialog = OkadaDialog

    @staticmethod
    def getRepresentedSource(sandbox):
        length = 5000.0
        return OkadaSource(
            length=length, width=15.0 * length**0.66, strike=45.0, rake=0, slip=2
        )

    def formatListItem(self):
        item = """
<span style="font-weight: bold; font-style: oblique">
    {idx}. {delegate.display_name}
    <span style="color: #616161;">
        ({delegate.display_backend})
    </span>
</span>
<table style="color: #616161; font-size: small;">
<tr>
    <td>Depth:</td><td>{source.depth:.2f} m</td>
</tr><tr>
    <td>Width:</td><td>{source.width:.2f} m</td>
</tr><tr>
    <td>Length:</td><td>{source.length:.2f} m</td>
</tr><tr>
    <td>Strike:</td><td>{source.strike:.2f}&deg;</td>
</tr><tr>
    <td>Dip:</td><td>{source.dip:.2f}&deg;</td>
</tr><tr>
    <td>Rake:</td><td>{source.rake:.2f}&deg;</td>
</tr><tr>
    <td>M<sub>0</sub>:</td><td>{source.seismic_moment:.2e}</td>
</tr><tr style="font-weight: bold;">
    <td>M<sub>W</sub>:</td><td>{source.moment_magnitude:.2f}</td>
</tr><tr style="font-weight: bold;">
    <td>Slip:</td><td>{source.slip:.2f} m</td>
</tr>
</table>
"""
        return item.format(idx=self.index.row() + 1, delegate=self, source=self.source)



import sys

from PyQt5 import QtCore, QtGui, QtWidgets

from kite.qt_utils import SceneLog, loadUi, validateFilename
from kite.sandbox_scene import SandboxScene

from .config import ConfigDialog
from .multiplot import ModelReferenceDockarea, SandboxSceneDockarea
from .sandbox_model import SandboxModel
from .sources_dock import SourcesListDock
from .tool_dialogs import ExtentDialog, LosDialog
from .util import get_resource


class Talpa(QtWidgets.QApplication):
    def __init__(self, filename=None):
        super().__init__(["Talpa"])

        splash_img = QtGui.QPixmap(get_resource("talpa_splash.png")).scaled(
            QtCore.QSize(400, 250), QtCore.Qt.KeepAspectRatio
        )
        self.splash = QtWidgets.QSplashScreen(
            splash_img, QtCore.Qt.WindowStaysOnTopHint
        )
        self.updateSplashMessage("Talpa")
        self.splash.show()
        self.processEvents()

        self.talpa_win = TalpaMainWindow(filename=filename)

        self.talpa_win.actionExit.triggered.connect(self.exit)
        self.aboutToQuit.connect(self.talpa_win.sandbox.worker_thread.quit)

        self.talpa_win.show()
        self.splash.finish(self.talpa_win)

    @QtCore.pyqtSlot(str)
    def updateSplashMessage(self, msg=""):
        self.splash.showMessage("Loading %s ..." % msg.title(), QtCore.Qt.AlignBottom)


class TalpaMainWindow(QtWidgets.QMainWindow):
    def __init__(self, *args, **kwargs):
        filename = kwargs.pop("filename", None)
        QtWidgets.QMainWindow.__init__(self, *args, **kwargs)
        loadUi(get_resource("talpa.ui"), baseinstance=self)
        self.sandbox = SandboxModel.empty(self)

        self.log = SceneLog(self, self.sandbox)

        self.actionSaveModel.triggered.connect(self.onSaveModel)
        self.actionLoadModel.triggered.connect(self.loadModel)
        self.actionExportKiteScene.triggered.connect(self.onExportScene)
        self.actionChangeExtent.triggered.connect(self.extentDialog)
        self.actionChangeLos.triggered.connect(self.losDialog)
        self.actionLoadReferenceScene.triggered.connect(self.onLoadReferenceScene)

        self.actionConfiguration.triggered.connect(self.configDialog)

        self.actionHelp.triggered.connect(
            lambda: QtGui.QDesktopServices.openUrl("https://pyrocko.org")
        )
        self.actionAbout_Talpa.triggered.connect(self.aboutDialog().show)
        self.actionLog.triggered.connect(self.log.show)

        self.sandbox.sigModelChanged.connect(self.createMisfitWindow)

        if filename is not None:
            self.loadModel(filename)
        self.createView(self.sandbox)

    def createView(self, sandbox):
        plots = SandboxSceneDockarea(sandbox, parent=self)
        sources = SourcesListDock(sandbox, parent=self)
        self.addDockWidget(QtCore.Qt.LeftDockWidgetArea, sources)
        self.centralwidget.layout().addWidget(plots)

    def aboutDialog(self):
        self._about = QtWidgets.QDialog(parent=self)
        loadUi(get_resource("about.ui"), baseinstance=self._about)
        return self._about

    @QtCore.pyqtSlot()
    def extentDialog(self):
        ExtentDialog(self.sandbox, self).show()

    @QtCore.pyqtSlot()
    def losDialog(self):
        LosDialog(self.sandbox, self).show()

    @QtCore.pyqtSlot()
    def configDialog(self):
        ConfigDialog(self).show()

    @QtCore.pyqtSlot()
    def onSaveModel(self):
        filename, _ = QtWidgets.QFileDialog.getSaveFileName(
            filter="YAML *.yml (*.yml)", caption="Save SandboxScene"
        )
        if not validateFilename(filename):
            return
        self.sandbox.model.save(filename)

    @QtCore.pyqtSlot()
    def loadModel(self, filename=None):
        if filename is None:
            filename, _ = QtWidgets.QFileDialog.getOpenFileName(
                filter="YAML *.yml (*.yml)", caption="Load SandboxScene"
            )
        if not validateFilename(filename):
            return
        model = SandboxScene.load(filename)
        self.sandbox.setModel(model)

    def onLoadReferenceScene(self):
        filename, _ = QtWidgets.QFileDialog.getOpenFileName(
            filter="YAML *.yml (*.yml)", caption="Load kite.Scene"
        )
        if not validateFilename(filename):
            return
        self.sandbox.model.loadReferenceScene(filename)

        self.createMisfitWindow()
        self.actionMisfitScene.setChecked(True)

    def createMisfitWindow(self):
        if self.sandbox.model.reference is None:
            return

        self.misfitWindow = MisfitWindow(self.sandbox, parent=self)

        def toggleWindow(switch):
            if switch:
                self.misfitWindow.show()
            else:
                self.misfitWindow.close()

        self.misfitWindow.windowClosed.connect(
            lambda: self.actionMisfitScene.setChecked(False)
        )
        self.actionMisfitScene.toggled.connect(toggleWindow)
        self.actionMisfitScene.setEnabled(True)

    def onExportScene(self):
        filename, _ = QtWidgets.QFileDialog.getSaveFileName(
            filter="YAML *.yml and NumPy container *.npz (*.yml *.npz)",
            caption="Save scene",
            parent=self,
        )
        if not validateFilename(filename):
            return
        scene = self.sandbox.model.getKiteScene()
        scene.save(filename)

    def closeModel(self, sandbox):
        pass


class MisfitWindow(QtWidgets.QMainWindow):
    windowClosed = QtCore.pyqtSignal()

    def __init__(self, sandbox, *args, **kwargs):
        QtWidgets.QMainWindow.__init__(self, *args, **kwargs)
        loadUi(get_resource("window_reference.ui"), self)

        self.move(
            self.parent().window().mapToGlobal(self.parent().window().rect().center())
            - self.mapToGlobal(self.rect().center())
        )

        self.sandbox = sandbox

        self.actionOptimizeSource.triggered.connect(self.sandbox.optimizeSource)

        self.createView(self.sandbox)

    def createView(self, sandbox):
        plots = ModelReferenceDockarea(sandbox)
        self.centralwidget.layout().addWidget(plots)

    def closeEvent(self, ev):
        self.windowClosed.emit()
        ev.accept()


def talpa(*args, **kwargs):
    talpa_app = Talpa(*args, **kwargs)
    talpa_app.exec()
    talpa_app.quit()



import matplotlib.pyplot as plt

from kite import Scene

# Assume we have a existing kite.Scene with defined quadtree parametrized
scene = Scene.load("acquila_2016.yml")

ax = plt.gca()
# Inspect the noise data which is used to calculate the covariance
ax.imshow(scene.covariance.noise_data)
plt.show()

# Inspect the focal-point (quick mode) covariance matrix
ax.imshow(scene.covariance.covariance_matrix_focal)

# Inspect the full covariance matrix
ax.imshow(scene.covariance.covariance_matrix)

# Get the full weight matrix
ax.imshow(scene.covariance.weight_matrix)

# Get the covariance and weight between two leafes
leaf_1 = scene.quadtree.leaves[0]
leaf_2 = scene.quadtree.leaves[0]

scene.covariance.getLeafCovariance(leaf_1, leaf_2)
scene.covariance.getLeafWeight(leaf_1, leaf_2)



import numpy as np
from pyrocko.guts import Tuple
from scipy import stats

from .plugin import Plugin, PluginConfig


class APSConfig(PluginConfig):
    patch_coords = Tuple.T(optional=True)
    model_coefficients = Tuple.T(optional=True)


class APS(Plugin):
    def __init__(self, scene, config=None):
        self.scene = scene
        self.config = config or APSConfig()
        self._log = scene._log.getChild("APS")

    def get_elevation(self):
        return self.scene.get_elevation()

    def set_patch_coords(self, xmin, ymin, xsize, ysize):
        self.config.patch_coords = (xmin, ymin, xsize, ysize)

    def get_patch_coords(self):
        if self.config.patch_coords is None:
            frame = self.scene.frame
            scene = self.scene
            rstate = np.random.RandomState(123)

            while True:
                llE = rstate.uniform(0, frame.lengthE * (4 / 5))
                llN = rstate.uniform(0, frame.lengthN * (4 / 5))
                urE = frame.lengthE / 5
                urN = frame.lengthN / 5

                colmin, colmax = frame.mapENMatrix(llE, llE + urE)
                rowmin, rowmax = frame.mapENMatrix(llN, llN + urN)

                displacement = scene._displacement[rowmin:rowmax, colmin:colmax]
                if np.any(displacement):
                    return llE, llN, urE, urN

        return self.config.patch_coords

    def get_data(self):
        scene = self.scene
        frame = self.scene.frame
        coords = self.get_patch_coords()
        if not coords:
            raise AttributeError("Set coordinates for APS.")

        colmin, colmax = frame.mapENMatrix(coords[0], coords[0] + coords[2])
        rowmin, rowmax = frame.mapENMatrix(coords[1], coords[1] + coords[3])
        displacement = scene._displacement[rowmin:rowmax, colmin:colmax]
        elevation = self.get_elevation()[rowmin:rowmax, colmin:colmax]

        mask = np.isfinite(displacement)
        elevation = elevation[mask]
        displacement = displacement[mask]

        return elevation, displacement

    def get_correlation(self):
        elevation, displacement = self.get_data()
        slope, intercept, _, _, _ = stats.linregress(
            elevation.ravel(), displacement.ravel()
        )

        return slope, intercept

    def apply(self, displacement):
        self._log.info("Applying APS model to displacement")

        scene = self.scene
        elevation = scene.get_elevation()
        slope, intercept = self.get_correlation()

        correction = elevation * slope + intercept
        displacement -= correction
        return displacement



import logging
import time

import numpy as np

"""Ellispoidal Cavity Model (ECM), triaxial elipsoidal deformation source.

Code is after Medhi Nikkhoo Matlab scripts found on
(http://www.volcanodeformation.com)

2017/05 - Marius Isken

Functions defined in this file serve as a backend for
kite.source.ellipsoidal_source

After

    Nikkhoo, M., Walter, T. R., Lundgren, P. R., Prats-Iraola, P. (2017):
    Compound dislocation models (CDMs) for volcano deformation analyses.
    Geophys Journal International, 208 (2): 877-894. doi:10.1093/gji/ggw427

    website:
    http://www.volcanodeformation.com
"""

logger = logging.getLogger("ECM")

d2r = np.pi / 180.0
r2d = 180.0 / np.pi
sqrt = np.sqrt
pi = np.pi


def cosd(deg):
    return np.cos(deg * d2r)


def sind(deg):
    return np.sin(deg * d2r)


def strike_dip(rot_mat, idx):
    Vstrike = np.array([-rot_mat[1, idx], rot_mat[0, idx], 0.0])
    Vstrike = Vstrike / np.linalg.norm(Vstrike)
    strike = np.arctan2(Vstrike[0], Vstrike[1]) * r2d
    if np.isnan(strike):
        strike = 0.0
    dip = np.arccos(rot_mat[2, idx]) * r2d
    return strike, dip


def rotation_matrix(rotx, roty, rotz):
    Rx = np.matrix(
        [[1.0, 0.0, 0.0], [0.0, cosd(rotx), sind(rotx)], [0.0, -sind(rotx), cosd(rotx)]]
    )
    Ry = np.matrix(
        [[cosd(roty), 0.0, -sind(roty)], [0.0, 1.0, 0.0], [sind(roty), 0.0, cosd(roty)]]
    )
    Rz = np.matrix(
        [[cosd(rotz), sind(rotz), 0.0], [-sind(rotz), cosd(rotz), 0.0], [0.0, 0.0, 1.0]]
    )

    return Rz * Ry * Rx


def pointCDM(coords, x0, y0, z0, rotx, roty, rotz, dVx, dVy, dVz, nu):
    """Point Compound Dislocation Model for surface displacements

    :param coords: Coordinates upon displacement is calculated
    :type coords: :class:`numpy.ndarray` of shape Nx2
    :param x0: Horizontal x-coordinate of the source, same unit as `coords`
    :type x0: float
    :param y0: Horizontal y-coordinate of the source, same unit as `coords`
    :type y0: float
    :param z0: Depth of the source, same unit as `coords`
    :type z0: float
    :param rotx: Clockwise rotation of ellipsoid around x-axis, [deg]
    :type rotx: float
    :param roty: Clockwise rotation of ellipsoid around y-axis, [deg]
    :type roty: float
    :param rotz: Clockwise rotation of ellipsoid around z-axis, [deg]
    :type rotz: float
    :param dVx: Volume change in axis-x, same unit as `coords`
    :type dVx: float
    :param dVy: Volume change in axis-y, same unit as `coords`
    :type dVy: float
    :param dVz: Volume change in axis-z, same unit as `coords`
    :type dVz: float
    :param nu: Poisson's ratio
    :type nu: float
    :returns: Volume change in axis-x
    :rtype: tuple of :class:`numpy.ndarray`
    """
    ncoords = coords.shape[0]
    rot_mat = rotation_matrix(rotx, roty, rotz)

    Ue = np.zeros(ncoords)
    Un = np.zeros(ncoords)
    Uv = np.zeros(ncoords)

    coords_shifted = coords.copy()
    coords_shifted[:, 0] -= x0
    coords_shifted[:, 1] -= y0

    component_names = ["dVx", "dVy", "dVz"]

    for icomp, comp in enumerate([dVx, dVy, dVz]):
        if np.all(comp):
            t0 = time.time()
            strike, dip = strike_dip(rot_mat, icomp)
            comp_ue, comp_un, comp_uv = PointDisplacementSurface(
                coords_shifted, z0, strike, dip, comp, nu
            )
            Ue += comp_ue
            Un += comp_un
            Uv += comp_uv
            logger.debug(
                "Calculated component %s [%.6f s]"
                % (component_names[icomp], time.time() - t0)
            )

    return Ue, Un, Uv


def ECM(coords, x0, y0, z0, rotx, roty, rotz, ax, ay, az, P, mu, lamda):
    """Calculate 2D surface displacement of a triaxial elipsoidal source.

    After:

        Nikkhoo, M., Walter, T. R., Lundgren, P. R., Prats-Iraola, P. (2017):
        Compound dislocation models (CDMs) for volcano deformation analyses.
        Geophys Journal International, 208 (2): 877-894. doi:10.1093/gji/ggw427

        website:
        http://www.volcanodeformation.com

    :param coords: Coordinates upon displacement is calculated
    :type coords: :class:`numpy.ndarray` of shape Nx2
    :param x0: Horizontal x-coordinate of the source, same unit as `coords`
    :type x0: float
    :param y0: Horizontal y-coordinate of the source, same unit as `coords`
    :type y0: float
    :param z0: Depth of the source, same unit as `coords`
    :type z0: float
    :param rotx: Clockwise rotation of ellipsoid around x-axis, [deg]
    :type rotx: float
    :param roty: Clockwise rotation of ellipsoid around y-axis, [deg]
    :type roty: float
    :param rotz: Clockwise rotation of ellipsoid around z-axis, [deg]
    :type rotz: float
    :param ax: Length of x semi-axis of ellisoid before rotation,
        same unit as `coords`
    :type ax: float
    :param ay: Length of y semi-axis of ellisoid before rotation,
        same unit as `coords`
    :type ay: float
    :param az: Length of z semi-axis of ellisoid before rotation,
        same unit as `coords`
    :type az: float
    :param P: Pressure on the cavity walls, same unit as the Lame constants
    :type P: float
    :param mu: Lame constant
    :type mu: float
    :param lamda: Lame constant
    :type lamda: float
    """
    ncoords = coords.shape[0]

    nu = lamda / (lamda + mu) / 2  # Poison's ratio
    K = lamda + 2 * mu / 3  # Bulk Modulus

    r0 = 1e-12  # Instability threshold for shape tensor
    ax = ax if ax > r0 else r0
    ay = ay if ay > r0 else r0
    az = az if az > r0 else r0

    a_arr = np.array([ax, ay, az])
    ia_sort = np.argsort(a_arr)[::-1]
    shape_tensor = shapeTensor(*a_arr[ia_sort], nu=nu)
    # Transform strain
    eT = -np.linalg.inv(shape_tensor) * P * np.ones((3, 1)) / 3.0 / K
    sT = (2 * mu * eT) + lamda * eT.sum()
    V = 4.0 / 3 * pi * ax * ay * az

    stress_tensor = sT[ia_sort]
    moment_tensor = V * stress_tensor

    dV = (eT.sum() - P / K) * V
    dV = dV + P * V / K  # Potency
    dVx, dVy, dVz = (
        1.0 / 2.0 / mu * (moment_tensor - lamda / 3.0 / K * moment_tensor.sum())
    )

    #####

    rot_mat = rotation_matrix(rotx, roty, rotz)

    Ue = np.zeros(ncoords)
    Un = np.zeros(ncoords)
    Uv = np.zeros(ncoords)

    coords_shifted = coords.copy()
    coords_shifted[:, 0] -= x0
    coords_shifted[:, 1] -= y0

    component_names = ["dVx", "dVy", "dVz"]
    for icomp, comp in enumerate([dVx, dVy, dVz]):
        if np.all(comp):
            t0 = time.time()
            strike, dip = strike_dip(rot_mat, icomp)
            comp_ue, comp_un, comp_uv = PointDisplacementSurface(
                coords_shifted, z0, strike, dip, float(comp), nu
            )
            Ue += comp_ue
            Un += comp_un
            Uv += comp_uv
            logger.debug(
                "Calculated component %s [%.6f s]"
                % (component_names[icomp], time.time() - t0)
            )

    return Ue, Un, Uv, dV, dV


def shapeTensor(a1, a2, a3, nu):
    """Calculates the Eshelby (1957) shape tensor components."""

    if a1 == 0.0 and a2 == 0.0 and a3 == 0:
        return np.zeros((3, 3)).view(np.matrix)

    # General case: triaxial ellipsoid
    if a1 > a2 and a2 > a3 and a3 > 0:
        logger.debug("General case: triaxial ellipsoid")
        sin_theta = sqrt(1 - a3**2 / a1**2)
        k = sqrt((a1**2 - a2**2) / (a1**2 - a3**2))

        # Calculate Legendre's incomplete elliptic integrals of the first and
        # second kind using Carlson (1995) method (see Numerical computation of
        # real or complex elliptic integrals. Carlson, B.C. Numerical
        # Algorithms (1995) 10: 13. doi:10.1007/BF02198293)
        tol = 1e-16
        c = 1 / sin_theta**2
        F = RF(c - 1, c - k**2, c, tol)
        E = F - k**2 / 3 * RD(c - 1, c - k**2, c, tol)

        I1 = (
            (4 * pi * a1 * a2)
            * a3
            / (a1**2 - a2**2)
            / sqrt(a1**2 - a3**2)
            * (F - E)
        )
        I3 = (
            (4 * pi * a1 * a2)
            * a3
            / (a2**2 - a3**2)
            / sqrt(a1**2 - a3**2)
            * (a2 * sqrt(a1**2 - a3**2) / a1 / a3 - E)
        )
        I2 = 4 * pi - I1 - I3

        I12 = (I2 - I1) / (a1**2 - a2**2)
        I13 = (I3 - I1) / (a1**2 - a3**2)
        I11 = (4 * pi / a1**2 - I12 - I13) / 3

        I23 = (I3 - I2) / (a2**2 - a3**2)
        I21 = I12
        I22 = (4 * pi / a2**2 - I23 - I21) / 3

        I31 = I13
        I32 = I23
        I33 = (4 * pi / a3**2 - I31 - I32) / 3

    # Special case-1: Oblate ellipsoid
    elif a1 == a2 and a2 > a3 and a3 > 0:
        logger.debug("Special case 1: oblate ellipsoid")
        I1 = (
            (2.0 * pi * a1 * a2)
            * a3
            / (a1**2 - a3**2) ** 1.5
            * (np.arccos(a3 / a1) - a3 / a1 * sqrt(1.0 - a3**2 / a1**2))
        )
        I2 = I1
        I3 = 4 * pi - 2 * I1

        I13 = (I3 - I1) / (a1**2 - a3**2)
        I11 = pi / a1**2 - I13 / 4
        I12 = I11

        I23 = I13
        I22 = pi / a2**2 - I23 / 4
        I21 = I12

        I31 = I13
        I32 = I23
        I33 = (4 * pi / a3**2 - 2 * I31) / 3

    # Special case-2: Prolate ellipsoid
    elif a1 > a2 and a2 == a3 and a3 > 0:
        logger.debug("Special case: prolate ellipsoid")
        I2 = (
            (2 * pi * a1 * a2)
            * a3
            / (a1**2 - a3**2) ** 1.5
            * (a1 / a3 * sqrt(a1**2 / a3**2 - 1) - np.arccosh(a1 / a3))
        )
        I3 = I2
        I1 = 4 * pi - 2 * I2

        I12 = (I2 - I1) / (a1**2 - a2**2)
        I13 = I12
        I11 = (4 * pi / a1**2 - 2 * I12) / 3

        I21 = I12
        I22 = pi / a2**2 - I21 / 4
        I23 = I22

        I32 = I23
        I31 = I13
        I33 = (4 * pi / a3**2 - I31 - I32) / 3

    # Special case-3: Sphere
    if a1 == a2 and a2 == a3:
        logger.debug("Special case: sphere")
        S1111 = (7.0 - 5 * nu) / 15.0 / (1.0 - nu)
        S1122 = (5 * nu - 1.0) / 15.0 / (1.0 - nu)
        S1133 = (5 * nu - 1.0) / 15.0 / (1.0 - nu)
        S2211 = (5 * nu - 1.0) / 15.0 / (1.0 - nu)
        S2222 = (7.0 - 5 * nu) / 15.0 / (1.0 - nu)
        S2233 = (5 * nu - 1.0) / 15.0 / (1.0 - nu)
        S3311 = (5 * nu - 1.0) / 15.0 / (1.0 - nu)
        S3322 = (5 * nu - 1.0) / 15.0 / (1.0 - nu)
        S3333 = (7.0 - 5 * nu) / 15.0 / (1.0 - nu)
    # General triaxial, oblate and prolate ellipsoids
    else:
        logger.debug("General case: triaxial, oblate and prolate ellipsoid")
        S1111 = (3.0 / 8.0 / pi) / (1.0 - nu) * (a1**2 * I11) + (
            1.0 - 2 * nu
        ) / 8.0 / pi / (1.0 - nu) * I1
        S1122 = (1.0 / 8.0 / pi) / (1.0 - nu) * (a2**2 * I12) - (
            1.0 - 2 * nu
        ) / 8.0 / pi / (1.0 - nu) * I1
        S1133 = (1.0 / 8.0 / pi) / (1.0 - nu) * (a3**2 * I13) - (
            1.0 - 2 * nu
        ) / 8.0 / pi / (1.0 - nu) * I1
        S2211 = (1.0 / 8.0 / pi) / (1.0 - nu) * (a1**2 * I21) - (
            1.0 - 2 * nu
        ) / 8.0 / pi / (1.0 - nu) * I2
        S2222 = (3.0 / 8.0 / pi) / (1.0 - nu) * (a2**2 * I22) + (
            1.0 - 2 * nu
        ) / 8.0 / pi / (1.0 - nu) * I2
        S2233 = (1.0 / 8.0 / pi) / (1.0 - nu) * (a3**2 * I23) - (
            1.0 - 2 * nu
        ) / 8.0 / pi / (1.0 - nu) * I2
        S3311 = (1.0 / 8.0 / pi) / (1.0 - nu) * (a1**2 * I31) - (
            1.0 - 2 * nu
        ) / 8.0 / pi / (1.0 - nu) * I3
        S3322 = (1.0 / 8.0 / pi) / (1.0 - nu) * (a2**2 * I32) - (
            1.0 - 2 * nu
        ) / 8.0 / pi / (1.0 - nu) * I3
        S3333 = (3.0 / 8.0 / pi) / (1.0 - nu) * (a3**2 * I33) + (
            1.0 - 2 * nu
        ) / 8.0 / pi / (1.0 - nu) * I3

    return np.matrix(
        [
            [S1111 - 1, S1122, S1133],
            [S2211, S2222 - 1, S2233],
            [S3311, S3322, S3333 - 1],
        ]
    )


def RF(x, y, z, r):
    """Calculates the RF term, Carlson (1995) method for elliptic integrals"""
    if x < 0 or y < 0 or z < 0:
        raise ArithmeticError("x, y and z values must be positive!")
    elif np.count_nonzero([x, y, z]) < 2:
        raise ArithmeticError("At most one of the x, y and z values can be zero!")

    xm = x
    ym = y
    zm = z
    A0 = (x + y + z) / 3
    Q = max([abs(A0 - x), abs(A0 - y), abs(A0 - z)]) / (3.0 * r) ** (1.0 / 6)
    n = 0
    Am = A0
    while abs(Am) <= Q / (4.0**n):
        lambdam = sqrt(xm * ym) + sqrt(xm * zm) + sqrt(ym * zm)
        Am = (Am + lambdam) / 4.0
        xm = (xm + lambdam) / 4.0
        ym = (ym + lambdam) / 4.0
        zm = (zm + lambdam) / 4.0
        n += 1
    X = (A0 - x) / 4**n / Am
    Y = (A0 - y) / 4**n / Am
    Z = -X - Y
    E2 = X * Y - Z**2
    E3 = X * Y * Z
    rf = (1.0 - E2 / 10 + E3 / 14 + E2**2.0 / 24 - 3.0 * E2 * E3 / 44.0) / sqrt(Am)
    return rf


def RD(x, y, z, r):
    """Calculates the RF term, Carlson (1995) method for elliptic integrals"""
    if z == 0:
        raise ArithmeticError("z value must be nonzero!")
    elif x == 0 and y == 0:
        raise ArithmeticError("At most one of the x and y values can be zero!")

    xm = x
    ym = y
    zm = z
    A0 = (x + y + 3 * z) / 5
    Q = max([abs(A0 - x), abs(A0 - y), abs(A0 - z)]) / (r / 4) ** (1.0 / 6)
    n = 0
    Am = A0
    S = 0
    while abs(Am) <= Q / (4**n):
        lambdam = sqrt(xm * ym) + sqrt(xm * zm) + sqrt(ym * zm)
        S = S + (1.0 / 4**n) / sqrt(zm) / (zm + lambdam)
        Am = (Am + lambdam) / 4
        xm = (xm + lambdam) / 4
        ym = (ym + lambdam) / 4
        zm = (zm + lambdam) / 4
        n += 1

    X = (A0 - x) / 4.0**n / Am
    Y = (A0 - y) / 4.0**n / Am
    Z = -(X + Y) / 3.0
    E2 = X * Y - 6 * Z**2
    E3 = (3 * X * Y - 8 * Z**2) * Z
    E4 = 3 * (X * Y - Z**2) * Z**2
    E5 = X * Y * Z**3
    rd = (
        1.0
        - 3 * E2 / 14
        + E3 / 6
        + 9 * E2**2 / 88
        - 3 * E4 / 22
        - 9 * E2 * E3 / 52
        + 3 * E5 / 26
    ) / 4**n / Am**1.5 + 3.0 * S
    return rd


def PointDisplacementSurface(coords_shifted, z0, strike, dip, dV, nu):
    """calculates surface displacements associated with a tensile
    point dislocation (PDF) in an elastic half-space (Okada, 1985).
    """
    ncoords = coords_shifted.shape[0]

    beta = strike - 90.0
    rot_mat = np.matrix([[cosd(beta), -sind(beta)], [sind(beta), cosd(beta)]])
    r_beta = rot_mat * coords_shifted.conj().T
    x = r_beta[0, :].view(np.ndarray).ravel()
    y = r_beta[1, :].view(np.ndarray).ravel()

    r = (x**2 + y**2 + z0**2) ** 0.5
    d = z0
    q = y * sind(dip) - d * cosd(dip)

    r3 = r**3
    rpd = r + d
    rpd2 = rpd**2
    a = (3 * r + d) / r3 / rpd**3

    I1 = (1.0 - 2 * nu) * y * (1.0 / r / rpd2 - x**2 * a)
    I3 = (1.0 - 2 * nu) * x / r3 - ((1.0 - 2 * nu) * x * (1.0 / r / rpd2 - y**2 * a))
    I5 = (1.0 - 2 * nu) * (1.0 / r / rpd - x**2 * (2 * r + d) / r3 / rpd2)

    # Note: For a PDF M0 = dV*mu!
    u = np.empty((ncoords, 3))

    u[:, 0] = x
    u[:, 1] = y
    u[:, 2] = d

    u *= (3.0 * q**2 / r**5)[:, np.newaxis]
    u[:, 0] -= I3 * sind(dip) ** 2
    u[:, 1] -= I1 * sind(dip) ** 2
    u[:, 2] -= I5 * sind(dip) ** 2
    u *= dV / 2 / pi

    r_beta = rot_mat.conj().T * u[:, :2].conj().T
    return (
        r_beta[0, :].view(np.ndarray).ravel(),  # ue
        r_beta[1, :].view(np.ndarray).ravel(),  # un
        u[:, 2],
    )  # uv


if __name__ == "__main__":
    nrows = 500
    ncols = 500

    x0 = 250.0
    y0 = 250.0
    depth = 30.0

    rotx = 0.0
    roty = 0.0
    rotz = 0.0

    ax = 1.0
    ay = 1.0
    az = 0.25

    # ax = 1.
    # ay = 1.
    # az = 1.

    P = 1e6
    mu = 0.33e11
    lamda = 0.33e11

    X, Y = np.meshgrid(np.arange(nrows), np.arange(ncols))

    coords = np.empty((nrows * ncols, 2))
    coords[:, 0] = X.ravel()
    coords[:, 1] = Y.ravel()

    ue, un, uv = ECM(coords, x0, y0, depth, rotx, roty, rotz, ax, ay, az, P, mu, lamda)



from .compound_models import EllipsoidSourceDelegate, PointCompoundSourceDelegate
from .okada import OkadaSourceDelegate
from .pyrocko import (
    PyrockoDoubleCoupleDelegate,
    PyrockoMomentTensorDelegate,
    PyrockoRectangularSourceDelegate,
    PyrockoRingfaultDelegate,
    PyrockoVLVDSourceDelegate,
)

__sources__ = [
    OkadaSourceDelegate,
    PyrockoRectangularSourceDelegate,
    PyrockoMomentTensorDelegate,
    PyrockoDoubleCoupleDelegate,
    PyrockoRingfaultDelegate,
    PyrockoVLVDSourceDelegate,
    EllipsoidSourceDelegate,
    PointCompoundSourceDelegate,
]



from datetime import datetime

import numpy as np

from .scene import BaseScene, Scene


def dtime(timestamp):
    return datetime.fromtimestamp(timestamp)


class TSScene(Scene):
    pass


class SceneStack(Scene):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._times = []
        self._scenes = {}

        self._selected_tmin = None
        self._selected_tmax = None

        self._range_tmin = None
        self._range_tmax = None

        self._scene_tmin = None
        self._scene_tmax = None

    @property
    def times(self):
        return sorted(self._scenes.keys())

    @property
    def tmin(self):
        return min(self.times)

    @property
    def tmax(self):
        return max(self.times)

    @property
    def timespan(self):
        return self.tmax - self.tmin

    @property
    def nscenes(self):
        return len(self._scenes)

    def add_scene(self, scene, timestamp):
        assert isinstance(scene, TSScene)
        if scene in self._scenes.values():
            raise AttributeError("Scene already in stack")

        self._log.info("Adding frame to stack at %s", dtime(timestamp))
        self._scenes[timestamp] = scene

        if self.nscenes == 1:
            self.phi = scene.phi
            self.theta = scene.theta
            self.rows, self.cols = scene.displacement.shape
            self.frame = scene.frame

        self.set_time_range(self.tmin, self.tmax)

    @property
    def displacement(self):
        displacement = self._scene_tmax.displacement - self._scene_tmin.displacement
        return displacement

    @property
    def displacement_mask(self):
        return ~np.isfinite(self.displacement)

    @displacement.setter
    def displacement(self):
        raise AttributeError("use add_frame to set the displacement")

    def get_scene_at(self, timestamp):
        times = self.times
        time = times[np.abs(np.array(times) - timestamp).argmin()]
        return time, self._scenes[time]

    def set_time_range(self, tmin, tmax):
        assert tmin <= tmax, "required tmin <= tmax"
        assert tmin >= self.tmin, "tmin outside of stack time range"
        assert tmax <= self.tmax, "tmax outside of stack time range"

        if self._range_tmin == tmin and self._range_tmax == tmax:
            self._log.debug("Time range unchanged")
            return

        self._log.info("Setting time range to %s - %s", dtime(tmin), dtime(tmax))

        self._selected_tmin = tmin
        self._selected_tmax = tmax

        self._range_tmin, scene_tmin = self.get_scene_at(tmin)
        self._range_tmax, scene_tmax = self.get_scene_at(tmax)

        if scene_tmin is self._scene_tmin and scene_tmax is self._scene_tmax:
            return

        self._scene_tmin = scene_tmin
        self._scene_tmax = scene_tmax

        self.evChanged.notify()



import time
from os import path as op

import numpy as np
from pyrocko import guts
from pyrocko.guts import Int, List, Object, String

from kite.scene import BaseScene, FrameConfig

# Import the modeling backends
from kite.sources import CompoundModelProcessor, DislocProcessor, PyrockoProcessor
from kite.util import Subject, property_cached

__processors__ = [DislocProcessor, PyrockoProcessor, CompoundModelProcessor]
km = 1e3
d2r = np.pi / 180.0
r2d = 180.0 / np.pi


class SandboxSceneConfig(Object):
    frame = FrameConfig.T(default=FrameConfig(), help="Frame/reference configuration")
    extent_north = Int.T(default=800, help="Model size towards north in [px]")
    extent_east = Int.T(default=800, help="Model size towards east in [px]")
    sources = List.T(help="List of sources")
    reference_scene = String.T(optional=True, help="Reference kite.Scene container")


class SandboxScene(BaseScene):
    def __init__(self, config=None, **kwargs):
        self.evChanged = Subject()
        self.evModelUpdated = Subject()
        self.evConfigChanged = Subject()
        self._initialised = False

        self.config = config if config else SandboxSceneConfig()
        super().__init__(frame_config=self.config.frame, **kwargs)

        self.reference = None
        self._los_factors = None

        for attr in ("theta", "phi"):
            data = kwargs.pop(attr, None)
            if data is not None:
                self.__setattr__(attr, data)

        self.setExtent(self.config.extent_east, self.config.extent_north)

        if self.config.reference_scene is not None:
            self.loadReferenceScene(self.config.reference_scene)

    @property
    def sources(self):
        """
        :returns: List of sources attached sandbox
        :rtype: list
        """
        return self.config.sources

    def setExtent(self, east, north):
        """Set the sandbox's extent in pixels

        :param east: Pixels in East
        :type east: int
        :param north: Pixels in North
        :type north: int
        """
        if self.reference is not None:
            self._log.warning("Cannot change a referenced model!")
            return

        self._log.debug("Changing model extent to %d px by %d px" % (east, north))

        self.cols = east
        self.rows = north

        self._north = np.zeros((self.rows, self.cols))
        self._east = np.zeros_like(self._north)
        self._down = np.zeros_like(self._north)

        self.theta = np.zeros_like(self._north)
        self.phi = np.zeros_like(self._north)
        self.theta.fill(np.pi / 2)
        self.phi.fill(0.0)

        self.config.extent_east = east
        self.config.extent_north = north

        self.frame.updateExtent()
        self._clearModel()
        self.processSources()

        self.evChanged.notify()

    def setLOS(self, phi, theta):
        """Set the sandbox's LOS vector

        :param phi: phi in degree
        :type phi: int
        :param theta: theta in degree
        :type theta: int
        """
        if self.reference is not None:
            self._log.warning("Cannot change a referenced model!")
            return

        self._log.debug("Changing model LOS to %d phi and %d theta", phi, theta)

        self.theta = np.full_like(self.theta, theta * r2d)
        self.phi = np.full_like(self.phi, phi * r2d)
        self.frame.updateExtent()

        self._clearModel()
        self.evChanged.notify()

    @property
    def north(self):
        if not self._initialised:
            self.processSources()
        return self._north

    @property
    def east(self):
        if not self._initialised:
            self.processSources()
        return self._east

    @property
    def down(self):
        if not self._initialised:
            self.processSources()
        return self._down

    @property_cached
    def displacement(self):
        """Displacement projected to LOS"""
        self.processSources()
        los_factors = self.los_rotation_factors

        self._displacement = (
            los_factors[:, :, 0] * -self._down
            + los_factors[:, :, 1] * self._east
            + los_factors[:, :, 2] * self._north
        )
        return self._displacement

    @property_cached
    def max_horizontal_displacement(self):
        """Maximum horizontal displacement"""
        return np.sqrt(self._north**2 + self._east**2).max()

    def addSource(self, source):
        """Add displacement source to sandbox

        :param source: Displacement Source
        :type source: :class:`kite.sources.meta.SandboxSource`
        """
        if source not in self.sources:
            self.sources.append(source)
        source._sandbox = self

        source.evParametersChanged.subscribe(self._clearModel)
        self._clearModel()

        self._log.debug("Source %s added" % source.__class__.__name__)

    def removeSource(self, source):
        """Remove displacement source from sandbox

        :param source: Displacement Source
        :type source: :class:`kite.sources.meta.SandboxSource`
        """
        source.evParametersChanged.unsubscribe(self._clearModel)
        self.sources.remove(source)
        self._log.debug("Source %s removed" % source.__class__.__name__)
        del source

        self._clearModel()

    def processSources(self):
        """Process displacement sources and update displacements"""
        result = self._process(self.sources)

        self._north += result["north"].reshape(self.rows, self.cols)
        self._east += result["east"].reshape(self.rows, self.cols)
        self._down += result["down"].reshape(self.rows, self.cols)
        self._initialised = True

    def processCustom(self, coordinates, sources, result_dict=None):
        return self._process(sources, result_dict)

    def _process(self, sources, result=None):
        if result is None:
            result = np.zeros(
                self.frame.npixel,
                dtype=[
                    ("north", float),
                    ("east", float),
                    ("down", float),
                ],
            )

        avail_processors = {}
        for proc in __processors__:
            avail_processors[proc.__implements__] = proc

        for impl in set([src.__implements__ for src in sources]):
            proc_sources = [
                src
                for src in sources
                if src.__implements__ == impl and src._cached_result is None
            ]

            if not proc_sources:
                continue

            processor = avail_processors.get(impl, None)

            if processor is None:
                self._log.warning("Could not find source processor for %s", impl)
                continue

            t0 = time.time()

            proc_result = processor(self).process(
                proc_sources, sandbox=self, nthreads=0
            )

            src_name = proc_sources[0].__class__.__name__
            self._log.debug(
                "Processed %s (nsources:%d) using %s [%.4f s]",
                src_name,
                len(proc_sources),
                processor.__name__,
                time.time() - t0,
            )

            result["north"] += proc_result["displacement.n"]
            result["east"] += proc_result["displacement.e"]
            result["down"] += proc_result["displacement.d"]

        return result

    def loadReferenceScene(self, filename):
        """Load a reference kite scene container into the sandbox

        A reference scene could be actually measured InSAR displacements.

        :param filename: filename of the scene container to load [.npy, .yml]
        :type filename: str
        """
        from .scene import Scene

        self._log.debug("Loading reference scene from %s", filename)
        scene = Scene.load(filename)
        self.setReferenceScene(scene)
        self.config.reference_scene = filename

    def setReferenceScene(self, scene):
        """Set a reference scene.

        A reference scene could be actually measured InSAR displacements.

        :param scene: Kite scene
        :type scene: :class:`kite.Scene`
        """
        self.config.frame = scene.config.frame

        self.setExtent(scene.cols, scene.rows)
        self.frame._updateConfig()

        self.phi = scene.phi
        self.theta = scene.theta
        self._los_factors = None
        self.reference = Reference(self, scene)
        self._log.debug("Reference scene set to scene.id:%s", scene.meta.scene_id)
        self.config.reference_scene = scene.meta.filename

        self._clearModel()

    def getKiteScene(self):
        """Return a :class:`kite.Scene` from current model.

        :returns: Scene
        :rtype: :class:`Scene`
        """
        from .scene import Scene, SceneConfig

        self._log.debug("Creating kite.Scene from SandboxScene")

        config = SceneConfig()
        config.frame = self.frame.config
        config.meta.scene_id = "Exported SandboxScene"

        return Scene(
            displacement=self.displacement,
            theta=self.theta,
            phi=self.phi,
            config=config,
        )

    def _clearModel(self):
        for arr in (self._north, self._east, self._down):
            arr.fill(0.0)
            if self.reference:
                arr[self.reference.scene.displacement_mask] = np.nan

        self.displacement = None
        self._los_factors = None
        self._initialised = False

        self.max_horizontal_displacement = None

        self.evModelUpdated.notify()

    def save(self, filename):
        """Save the sandbox as kite scene container

        :param filename: filename to save under
        :type filename: str
        """
        _file, ext = op.splitext(filename)
        filename = filename if ext in [".yml"] else filename + ".yml"
        self._log.debug("Saving model scene to %s" % filename)
        for source in self.sources:
            source.regularize()
        self.config.dump(
            filename="%s" % filename, header="kite.SandboxScene YAML Config"
        )

    @classmethod
    def load(cls, filename):
        """Load a :class:`kite.SandboxScene`

        :param filename: Config file to load [.yml]
        :type filename: str
        :returns: A sandbox from config file
        :rtype: :class:`kite.SandboxScene`
        """
        config = guts.load(filename=filename)
        sandbox_scene = cls(config=config)
        sandbox_scene._log.debug("Loading config from %s" % filename)
        for source in sandbox_scene.sources:
            sandbox_scene.addSource(source)
        return sandbox_scene


class Reference(object):
    def __init__(self, model, scene):
        self.model = model
        self.scene = scene

        self.model.evModelUpdated.subscribe(self._clearRefence)

    def _clearRefence(self):
        self.difference = None

    @property_cached
    def difference(self):
        return self.scene.displacement - self.model.displacement

    def optimizeSource(self, callback=None):
        from scipy import optimize

        quadtree = self.scene.quadtree
        coordinates = quadtree.leaf_coordinates
        sources = self.model.sources

        model_result = np.zeros(
            coordinates.shape[0],
            dtype=[
                ("north", float),
                ("east", float),
                ("down", float),
            ],
        )

        if len(sources) > 1 or not sources:
            self._log.warning("We can optimize single, individual sources only!")
            return
        source = sources[0]
        source.evParametersChanged.mute()

        def misfit(model_displacement, lp_norm=2):
            p = lp_norm
            mf = (
                np.sum(np.abs(quadtree.leaf_medians - model_displacement) ** p) ** 1.0
                / p
            )
            return mf

        def kernel(model):
            source.setParametersArray(model)
            res = self.model.processCustom(coordinates, [source], model_result)

            model_displacement = (
                quadtree.leaf_los_rotation_factors[:, 0] * -res["down"]
                + quadtree.leaf_los_rotation_factors[:, 1] * res["east"]
                + quadtree.leaf_los_rotation_factors[:, 2] * res["north"]
            )
            mf = misfit(model_displacement)

            model_result.fill(0.0)
            return mf

        result = optimize.minimize(
            kernel,
            source.getParametersArray(),
            method="SLSQP",
            bounds=None,
            constraints=(),
            tol=None,
            callback=callback,
            options={
                "disp": True,
                "iprint": 10,
                "eps": 1.4901161193847656e-04,
                "maxiter": 300,
                "ftol": 1e-06,
            },
        )

        source.evParametersChanged.unmute()
        self.model.sources[0].setParametersArray(result.x)
        return result


class TestSandboxScene(SandboxScene):
    @classmethod
    def randomOkada(cls, nsources=1):
        from .sources import OkadaSource

        sandbox_scene = cls()

        def r(lo, hi):
            return float(np.random.randint(lo, high=hi, size=1))

        for isrc in range(nsources):
            length = r(5000, 15000)
            sandbox_scene.addSource(
                OkadaSource(
                    easting=r(sandbox_scene.frame.E.min(), sandbox_scene.frame.E.max()),
                    northing=r(
                        sandbox_scene.frame.N.min(), sandbox_scene.frame.N.max()
                    ),
                    depth=r(0, 8000),
                    strike=r(0, 360),
                    dip=r(0, 90),
                    slip=r(1, 5),
                    rake=r(-180, 180),
                    length=length,
                    width=15.0 * length**0.66,
                )
            )
        return sandbox_scene

    @classmethod
    def simpleOkada(cls, **kwargs):
        from .sources import OkadaSource

        sandbox_scene = cls()

        parameters = {
            "easting": 50000,
            "northing": 50000,
            "depth": 0,
            "strike": 180.0,
            "dip": 20,
            "slip": 2,
            "rake": 90,
            "length": 10000,
            "width": 15.0 * 10000**0.66,
        }
        parameters.update(kwargs)

        sandbox_scene.addSource(OkadaSource(**parameters))
        return sandbox_scene
